// ChooseTargetDialog.java created on 07 March 2004 at 11:21

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.table.*;
import javax.swing.border.*;
import java.io.*;

import kea.programs.dialogBoxes.databaseDialog.*;
import kea.database.*;
import kea.programs.dialogBoxes.*;
import kea.io.*;
import kea.util.*;
import kea.graphics.Colour;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Dialog box for choosing target group.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public class ChooseTargetDialog extends AbstractSearchDatabaseDialog {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS

	private JTable _table;
	private ChooseTargetDialogTableModel _model;
	private Database _database;
	private MainClass _owner;
	private File _resultsFile;
	private JMenu _helpMenu;
	private JToolBar _toolBar;
	
	private boolean _rootVisible;
	private RootVisibleAction _rootVisibleAction;
	
	private DialogResult _dialogResult;
	private JTree _tree;
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of ChooseTargetDialog.
	 @param owner JPrimrose MainClass.
	 @param database Database object containing data to display.
	 */
	public ChooseTargetDialog(MainClass owner, Database database) {
		
		super(
			owner, 
			database, 
			new ChooseTargetDialogTreeCellRenderer(),
			false
			);
		
		_database = database;
		_owner = owner;
		
		_dialogResult = DialogResult.None;
		
		// Root visible?
		_rootVisible = false;
		this.getTree().setRootVisible(_rootVisible);
		_rootVisibleAction = new RootVisibleAction();
		
		// Add table and graph to dialog.
		JSplitPane splitPane = new JSplitPane(
			JSplitPane.VERTICAL_SPLIT,
			this.tableComponent(), 
			this.infoPanelComponent()
			);
		splitPane.setDividerLocation(160);
		this.addRightComponent(splitPane);		
		
		// Add menu.
		this.setJMenuBar(menuBarComponent());
		
		// Add tool bar.
		this.addToolBar(toolBarComponent());
		this.addButtonPanel(buttonPanelComponent());
		
		_tree = this.getTree();
		
		_tree.setCellRenderer(new ChooseTargetDialogTreeCellRenderer());
		
		// Set up Tree listeners to update table.
		_tree.addTreeSelectionListener(new SelectionListener());
		_tree.addTreeExpansionListener(new ExpansionListener());
		
		// Add tree mouse listener to change target status of records.
		_tree.addMouseListener(new ChooseTargetDialog.TreeMouseListener());
		
		this.enableTreeRecordSelection(false);
		
		} // End of constructor.
	
	//##########################################################################
	
	// PROPERTY ACCESSOR METHODS
	
	//##########################################################################
	
	// METHODS
	
	/**
	 Displays database dialog.
	 @return Return value depending on database dialog button pressed.
	 */
	public DialogResult showDatabaseDialog() {
	
		this.show();
		return _dialogResult;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JToolBar toolBarComponent() {
		
		_toolBar = new JToolBar();
		_toolBar.setRollover(true);
		_toolBar.add(_rootVisibleAction);
		_toolBar.addSeparator();
		
		return _toolBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel buttonPanelComponent() {
		
		JPanel panel = new JPanel();
		panel.setBorder(new EmptyBorder(5, 5, 5, 5));
		panel.setLayout(new GridLayout(1, 3, 5, 5));
		
		JButton okButton = new JButton(new OkAction());
		panel.add(okButton);
		okButton.setPreferredSize(new Dimension(100, 30));
		
		JButton cancelButton = new JButton(new CancelAction());
		panel.add(cancelButton);
		
		JButton resetButton = new JButton(new ResetAction());
		panel.add(resetButton);
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JMenuBar menuBarComponent() {
		
		JMenuBar menuBar = new JMenuBar();
		menuBar.setBorder(new EmptyBorder(0,0,0,0));
		
		// File.
		JMenu fileMenu = new JMenu("File");
		menuBar.add(fileMenu);
		fileMenu.getAccessibleContext().setAccessibleName("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		fileMenu.add(new ExitAction());
		
		// Options.
		JMenu optionsMenu = new JMenu("Options");
		menuBar.add(optionsMenu);
		optionsMenu.getAccessibleContext().setAccessibleName("Options");
		optionsMenu.setMnemonic(KeyEvent.VK_O);
		optionsMenu.add(_rootVisibleAction);
		
		// Help.
		_helpMenu = new JMenu("Help");
		menuBar.add(_helpMenu);
		_helpMenu.getAccessibleContext().setAccessibleName("Help");
		_helpMenu.setMnemonic(KeyEvent.VK_H);
		
		return menuBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Use this method to add menu items (Actions) to the help menu.
	 @param action Action to add.
	 */
	public void addHelpMenuAction(AbstractAction action) {
		_helpMenu.add(action);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JComponent infoPanelComponent() {
		
		return new JPanel();
		
		} // End of method.

	////////////////////////////////////////////////////////////////////////////
	
	private JComponent tableComponent() {

		_model = new ChooseTargetDialogTableModel(
			RecordFactory.createDummyRecord()
			);
		
		// Initialise TableCellRenderer.
		TableCellRenderer renderer = new ChooseTargetDialogTableCellRenderer();
		
		// Initialise table with TableModel.
		_table = new JTable(_model);
		
		// Add TableCellRenderer for dealing with name and icon objects.
		_table.setDefaultRenderer(Record.class, renderer);
		
		// Determine how columns should be resized.
		_table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
		
		// Set column widths.
		_model.setColumnWidths(_table);
		
		// Do not display grid.
		_table.setShowGrid(false);
		
		// Set selection colours.
		//_table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		_table.setSelectionBackground(Colour.BackgroundSelection);
		_table.setSelectionForeground(Color.BLACK);
		
		// Add mouse listener for record selection/deselection.
		_table.addMouseListener(new TableMouseListener());
		
		// Add to scroll pane, set background to white and return.
		JScrollPane sp = new JScrollPane(_table);
		sp.getViewport().setBackground(Color.WHITE);
		return sp;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Fills table with data supplied by Record object.  
	private void updateTable(Record record) {
		
		// Do nothing if a leaf.
		if (record.isLeaf()) {return;}
		
		// Get model associated with table.
		ChooseTargetDialogTableModel m = 
			(ChooseTargetDialogTableModel)_table.getModel();
			
		// Display new data.
		m.updateModel(record);

		// Reset column sizes (otherwise all same size).
		m.setColumnWidths(_table);

		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Inherited method - updates table when method called.
	 @param record Record object currently displayed.
	 */
	public void fireOnTreeNodeSelectionToggle(Record record) {
		
		this.updateTable(_model.getCurrentRecord());
		
		}
	
	////////////////////////////////////////////////////////////////////////////
	
	// Ensures expanded status of tree is up-to-date with table.  When a table
	// record is double clicked, and the children of that record displayed in 
	// the table, so should the tree be expanded to reflect this change.
	private void expandTree(Record record) {
		
		// Procedure: Get all parents of record.  Create an array from these
		// Record objects.  Use this array to create a TreePath object.  Use
		// this TreePath object to expand tree.
		
		// Stack to store record and parents of record.
		Stack stack = new Stack();
		stack.push(record);
		
		// Store parents in stack (order: most recent parent to root).
		Record parent = null;
		while ((parent = (Record)record.getParent()) != null) {
			stack.push(parent);
			record = parent;
			}
		
		// Now create array and fill it with stack contents in reverse order. 
		int size = stack.size();
		Record[] parents = new Record[size];
		
		int counter = 0;
		while(!stack.empty()) {
			parents[counter] = (Record)stack.pop();
			counter++;
			}
		
		// Create TreePath object and use to expand tree.
		TreePath tp = new TreePath(parents);
		this.getTree().expandPath(tp);
		
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	// Open record within table and expand tree.
	private class TableMouseListener extends MouseAdapter {
		
		public void mouseClicked(MouseEvent me) {
			if (me.getClickCount() == 2) {openRecord(me);}
			if (me.isPopupTrigger()) {targetRecordToggle(me);}
			}
		
		public void mouseEntered(MouseEvent me) {
			}
		
		public void mouseExited(MouseEvent me) {
			}
		
		public void mousePressed(MouseEvent me) {
			if (me.isPopupTrigger()) {targetRecordToggle(me);}
			}
		
		public void mouseReleased(MouseEvent me) {
			if (me.isPopupTrigger()) {targetRecordToggle(me);}
			}
		
		
		//=====================================================
		
		private void targetRecordToggle(MouseEvent me) {
			
			int rowIndex = _table.rowAtPoint(me.getPoint());
			int columnIndex = _table.columnAtPoint(me.getPoint());
			
			Object o = _table.getModel().getValueAt(rowIndex, columnIndex);
			
			if (o instanceof Record) {
				Record r = (Record)o;
				
				toggleSelection(r);
				
				}
			}
		
		private void openRecord(MouseEvent me) {
			
			int rowIndex = _table.rowAtPoint(me.getPoint());
			int columnIndex = _table.columnAtPoint(me.getPoint());
			
			Object o = _table.getModel().getValueAt(rowIndex, columnIndex);
			
			if (o instanceof Record) {
				Record r = (Record)o;
				updateTable(r);
				expandTree(r);
				//_graphUpdater.updateWith(r);
				}
			
			} 
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Marks records as targeted or not targeted.
	private class TreeMouseListener implements MouseListener {
		
		public void mouseClicked(MouseEvent me) {
			if (me.isPopupTrigger()) {targetRecordToggle(me);}
			}
		
		public void mouseEntered(MouseEvent me) {
			}
		
		public void mouseExited(MouseEvent me) {
			}
		
		public void mousePressed(MouseEvent me) {
			if (me.isPopupTrigger()) {targetRecordToggle(me);}
			}
		
		public void mouseReleased(MouseEvent me) {
			if (me.isPopupTrigger()) {targetRecordToggle(me);}
			}
		
		//=====================================================
		
		private void targetRecordToggle(MouseEvent me) {
			
			TreePath tp = _tree.getClosestPathForLocation(me.getX(), me.getY());
			if (tp == null) {return;}
			Record r = (Record)tp.getLastPathComponent();
			toggleSelection(r);
			
			} 
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Update table when tree changes.
	private class SelectionListener implements TreeSelectionListener {
		
		public void valueChanged(TreeSelectionEvent e) {
			Record r = (Record)e.getPath().getLastPathComponent();
			
			updateTable(r);
			//_graphUpdater.updateWith(r);
			
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Update table when tree changes.
	private class ExpansionListener implements TreeExpansionListener {
		
		public void treeCollapsed(TreeExpansionEvent e) {
			
			}
		
		public void treeExpanded(TreeExpansionEvent e) {
			Record r = (Record)e.getPath().getLastPathComponent();
			
			updateTable(r);
			
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class RootVisibleAction extends AbstractAction {
		
		public RootVisibleAction() {
			putValue(Action.NAME, "Show Root");
			putValue(Action.SHORT_DESCRIPTION, "Displays root icon at base of tree.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_R));
			putValue(
				Action.ACCELERATOR_KEY,
				KeyStroke.getKeyStroke(KeyEvent.VK_R, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(
				Action.SMALL_ICON, 
				kea.programs.dialogBoxes.databaseDialog.Info.ROOT_VISIBLE_ICON
				);
			} 
		
		public void actionPerformed(ActionEvent e) {
			_rootVisible = !_rootVisible;
			ChooseTargetDialog.this.getTree().setRootVisible(_rootVisible);
			}		
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Exit.
	private class ExitAction extends AbstractAction {
		public ExitAction() {
			putValue(Action.NAME, "Exit Results");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Exit Results dialog and returns to main window."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_X));
			putValue(Action.SMALL_ICON, Info.getExitIcon());
			} 
		public void actionPerformed(ActionEvent ae) {
			WindowEvent we = new WindowEvent(
				ChooseTargetDialog.this, 
				WindowEvent.WINDOW_CLOSING
				);
			processWindowEvent(we);
			} 
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class OkAction extends AbstractAction {
		public OkAction() {
			putValue(Action.NAME, "Accept");
			putValue(Action.SHORT_DESCRIPTION, "Accept changes made.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_A));
			} 
		
		public void actionPerformed(ActionEvent e) { 
			_dialogResult = DialogResult.Ok;
			}		
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class CancelAction extends AbstractAction {
		
		public CancelAction() {
			putValue(Action.NAME, "Cancel");
			putValue(Action.SHORT_DESCRIPTION, "Cancel changes made.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_C));
			}
		
		public void actionPerformed(ActionEvent e) {
			_dialogResult = DialogResult.Cancel;
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class ResetAction extends AbstractAction {
		
		public ResetAction() {
			putValue(Action.NAME, "Reset");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Resets all records to non-target status."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_R));
			} 
		
		public void actionPerformed(ActionEvent e) {
			_dialogResult = DialogResult.Reset;
			}
		
		} // End of inner class.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
