// FindOligosDialog.java created on 13 March 2004 at 14:43

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.io.*;
import java.awt.*;
import java.lang.*;
import java.text.*;
import java.util.*;
import javax.swing.*;
import java.awt.event.*;
import javax.swing.border.*;
import javax.swing.table.*;
import javax.swing.event.*;

import kea.programs.icons.KeaIcon;
import kea.programs.progress.ProgressComponent;
import kea.database.*;
import kea.util.GUIUtils;
import kea.programs.dialogBoxes.databaseDialog.*;
import kea.graphics.Colour;
import kea.programs.buttons.*;
import kea.io.*;
import kea.exceptions.*;
import kea.parsers.fasta.*;
import kea.parsers.genBank.*;
import kea.event.*;
import kea.algorithm.alignment.*;
import kea.sequences.nucleicAcid.*;
import kea.programs.dialogBoxes.*;
import kea.programs.table.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Dialog for designing oligos.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public class FindOligosDialog extends JDialog {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	private static final int MAX_NUMBER_OF_TARGETS = 500;
	private static final int PREFERRED_MAX_NUMBER_OF_TARGETS = 15;
	
	private static final int DEFAULT_OLIGO_LENGTH = 16;
	private static final int MAXIMUM_OLIGO_LENGTH = 100;
	private static final int MINIMUM_OLIGO_LENGTH = 3;
	private static final int DEFAULT_NUMBER_OF_DEGENERATE_BASES = 3;
	private static final int MAXIMUM_NUMBER_OF_DEGENERATE_BASES = 10;
	
	private static final int NON_ALIGNMENT_CHARACTER = 
		CharacterType.ConsensusNonAligned.toChar();
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private final TargetCollection _targetCollection;
	
	private JMenu _helpMenu;
	private final Database _database;
	private boolean _abortParser;
	private ProgressComponent _progressComponent;
	private ArrayList _sequenceBuffer; // Temp storage prior to consensus generation.
	private JTextArea _resultsTextArea;
	private JSpinner _oligoLengthSpinner;
	private JSpinner _degenerateBaseSpinner;
	private DialogResult _dialogResult;
	
	private Action _acceptAction;
	private Action _resetAction;
	private Action _cancelAction;
	private Action _findAction;
	
	private KeaButton _findButton;
	
	private PrimroseOligoDatabase _oligoDatabase;
	
	private FindOligosDialogTableModel _model;
	
	private OligoDesignInfo _oligoDesignInfo;
	
	private final Action _notepadAction;
	
	private final int _initialOligoLength;
	private final int _initialDegeneracy;
	
	private int _oligoLengthUsedToGenerateOligos;
	private int _degeneracyUsedToGenerateOligos;
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of FindOligosDialog.
	 @param owner JFrame owner of dialog box.
	 @param database Database object associated with dialog.
	 */
	public FindOligosDialog(
		final JFrame owner, 
		final Database database,
		final TargetCollection targetCollection,
		final PrimroseOligoDatabase oligoDatabase,
		final OligoDesignInfo oligoDesignInfo,
		final Action notepadAction
		) {	
			
		super(owner, "Find potentially useful oligonucleotides", true);	
			
		_notepadAction = notepadAction;
		
		_oligoDesignInfo = oligoDesignInfo;
		_oligoDatabase = oligoDatabase;
		if (_oligoDesignInfo != null) {
			displayResultsSummary(_oligoDesignInfo.getMaximumDegeneracy());
			
			_initialOligoLength = _oligoDesignInfo.getOligoLength();
			_initialDegeneracy = _oligoDesignInfo.getMaximumDegeneracy();
			
			}
		else {
			_initialOligoLength = DEFAULT_OLIGO_LENGTH;
			_initialDegeneracy = DEFAULT_NUMBER_OF_DEGENERATE_BASES;
			}
	
		_database = database;
		
		// If no initialised target collection has been passed to object, create
		// new target collection from database object.
		if (targetCollection == null) {
			_targetCollection = getTargets(_database.getRecords());
			}
		// If an existing collection has been passed to object, create new target
		// collection from database object then merge with existing collection
		// (i.e., taking into account any previously selected targets).
		else {
			_targetCollection = merge(
				targetCollection, 
				getTargets(_database.getRecords())
				);
			}
		
		
		// Initialise array which will contain target sequences.
		_sequenceBuffer = new ArrayList();
		
		_abortParser = false;
		
		initialise();
		
		pack();
		GUIUtils.setLocationOnceSized(this);
		GUIUtils.setWindowMinimumSize(this, this.getSize());
		
		
		
		
		} // End of constructor.
	
	//##########################################################################
	
	// METHODS
	
	// Merge new collection with old.
	private TargetCollection merge(
		final TargetCollection oldCollection, 
		final TargetCollection newCollection
		) {
		
		TargetCollection merged = new TargetCollection();	
			
		for (int i = 0; i < newCollection.getSize(); i++) {
			
			Target newTarget = newCollection.getTarget(i);
			Target oldTarget = oldCollection.getTarget(newTarget.getRecord());
			
			// Equivalent target absent from old collection - store new.
			if (oldTarget == null) {
				merged.addTarget(newTarget);
				}
			
			// Equivalent target present in old collection - clone old and store.
			else {
				merged.addTarget(Target.clone(oldTarget));
				}
			}	
			
		return merged;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void initialise() {
		
		JPanel mainPanel = (JPanel)this.getContentPane();
		mainPanel.setLayout(new BorderLayout());
		
		// Menu bar.
		this.setJMenuBar(initialiseMenuBarComponent());
		
		JPanel innerPanel1 = new JPanel(new BorderLayout());
		JPanel innerPanel2 = new JPanel(new BorderLayout());
		
		// main panel north - toolbar.
		mainPanel.add(initialiseToolBar(), BorderLayout.NORTH);
		
		// Main panel center - inner panel 1.
		mainPanel.add(innerPanel1, BorderLayout.CENTER);
		 
		// Main panel south - progress bar.
		mainPanel.add(initialiseProgressBarComponent(), BorderLayout.SOUTH);
		
		
		// Inner panel 1 north - empty.
		
		// Inner panel 1 center - table.
		innerPanel1.add(initialiseTableComponent(), BorderLayout.CENTER);
		
		// Inner panel 1 south - inner panel 2.
		innerPanel1.add(innerPanel2, BorderLayout.SOUTH);
	
		
		// Inner panel 2 north - oligo parameters + find button.
		innerPanel2.add(initialiseParameterFieldsComponent(), BorderLayout.NORTH);
		
		// Inner panel 2 center - results panel.
		innerPanel2.add(initialiseResultsDisplayComponent(), BorderLayout.CENTER);
		
		// Inner panel 2 south - Buttons.
		innerPanel2.add(initialiseButtonComponent(), BorderLayout.SOUTH);

		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JToolBar initialiseToolBar() {
		
		JToolBar toolBar;
		
		toolBar = new JToolBar();
		toolBar.setRollover(true);
		toolBar.add(_notepadAction);
		toolBar.addSeparator();
		
		return toolBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns OligoDatabase object containing separate OligoCollection objects.
	 Each OligoCollection contains Oligo objects with a specific number of
	 degenerate bases (corresponding to the index number of the OligoCollection).
	 @return Oligo database object.
	 */
	public PrimroseOligoDatabase getOligoDatabase() {
		
		// Just in case method called too early.
		if (_oligoDatabase == null) {
			throw new RuntimeException(
				"Oligo database has not been filled yet!"
				);
			}
		
		return _oligoDatabase;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns collection of target objects, with those selected for use in 
	 oligo design noted.
	 @return TargetCollection object.
	 */
	public TargetCollection getTargetCollection() {
		return _targetCollection;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns user defined oligo design info for display later in program.
	 @return OligoDesignInfor object.
	 */
	public OligoDesignInfo getOligoDesignInfo() {
		return _oligoDesignInfo;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Displays dialog.
	 @return Returns DialogResult enum according to button pressed by user.
	 */
	public DialogResult showDialog() {
		
		setVisible(true);
		return _dialogResult;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JMenuBar initialiseMenuBarComponent() {
		
		JMenuBar menuBar = new JMenuBar();
		menuBar.setBorder(new EmptyBorder(0,0,0,0));
		
		// Help.
		_helpMenu = new JMenu("Help");
		menuBar.add(_helpMenu);
		_helpMenu.getAccessibleContext().setAccessibleName("Help");
		_helpMenu.setMnemonic(KeyEvent.VK_H);
		
		return menuBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////	
		
	/**
	 Use this method to add menu items (Actions) to the help menu.
	 @param action Action to add.
	 */
	public void addHelpMenuAction(final AbstractAction action) {
		_helpMenu.add(action);
		} // End of method.
		
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseTableComponent() {
		
		JPanel panel = new JPanel(new GridLayout(1, 1));
		panel.setBorder(
			new CompoundBorder(
				new EmptyBorder(5, 5, 5, 5),
				new TitledBorder("Records to target")
				)
			);
		
		// Initialise TableModel with target data.
		_model = new FindOligosDialogTableModel(_targetCollection);
		
		// Initialise table with TableModel.
		//JTable table = new JTable(_model);
		
		// Enable sorting - uses classes from Java tutorial.
		TableSorter sorter = new TableSorter(_model); 
		JTable table = new JTable(sorter);           
		sorter.addMouseListenerToHeaderInTable(table); 

		// Add TableCellRenderer for dealing with name and icon objects.
		table.setDefaultRenderer(Record.class, new FindOligoDialogTableCellRenderer());
		table.setDefaultRenderer(Integer.class, new IntegerTableCellRenderer());
		table.setDefaultRenderer(Target.class, new BooleanTableCellRenderer());  
		
		//TableColumnModel tcm = table.getColumnModel();
		//TableColumn tc = tcm.getColumn(_model.CHECKBOX_COLUMN);
		//tc.setCellEditor(new BooleanTableCellEditor());
		
		
		// Set column sizes
		TableColumn column = null;
		for (int i = 0; i < _model.getColumnCount(); i++) {
			column = table.getColumnModel().getColumn(i);
			if (i == _model.NUMBER_COLUMN) { // Number
				column.setPreferredWidth(5); 	
				} 
			else if (i == _model.NAME_COLUMN) {
				column.setPreferredWidth(180); 	
				}
			else if (i == _model.CHECKBOX_COLUMN) {
				column.setPreferredWidth(60);
				}
			else {
				column.setPreferredWidth(50);
				}
			}
		
		// Determine how columns should be resized.
		table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);

		// Do not display grid.
		table.setShowGrid(false);
		
		// Set selection colours.
		table.setSelectionBackground(Colour.BackgroundSelection);
		table.setSelectionForeground(Color.BLACK);
		//table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		
		ListSelectionModel lsm = table.getSelectionModel();
		
		// Add listener to listen for selection status changes to targets.
		lsm.addListSelectionListener(new MyListSelectionListener());
		
		// Add to scroll pane, set background to white and return.
		JScrollPane sp = new JScrollPane(table);
		sp.getViewport().setBackground(Color.WHITE);
		
		panel.add(sp);
		
		panel.setPreferredSize(new Dimension(500, 150)); // <+++++++++++++++++++++++ KEY SIZING STEP!
		
		return panel;
	
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Collects records together.
	private TargetCollection getTargets(final Record[] records) {
		
		TargetCollection collection = new TargetCollection();

		for (int i = 0; i < records.length; i++) {
			if (records[i].getRecordType() == RecordType.Target) {
				collection.addTarget(
					new Target(
						records[i], 
						false // By default, targets are not selected for use.
						)
					);
				}
			}

		return collection;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Packages those target records selected by user to design oligos, for use
	// by appropriate database file parsers.
	private Record getTargetsToFindOligosFrom(
		final TargetCollection targetCollection,
		final DatabaseFile file
		) {
		
		Record targets = RecordFactory.createEmptyRecord();	
			
		for (int i = 0; i < targetCollection.getSize(); i++) {
			Target t = targetCollection.getTarget(i);
			
			// Add to Record object if selected by user AND source file
			// corresponds to method argument.
			if (t.isSelected() && t.getRecord().getSource() == file) {	
				targets.addChild(
					t.getRecord()
					);
				}
			}	
			
		return targets;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseResultsDisplayComponent() {
		
		JPanel panel = new JPanel(new GridLayout(1, 1));
		panel.setBorder(
			new CompoundBorder(
				new EmptyBorder(5, 5, 5, 5),
				new TitledBorder("Results summary")
				)
			);
		
		_resultsTextArea = new JTextArea();
		panel.add(new JScrollPane(_resultsTextArea));
		_resultsTextArea.setLineWrap(true);
		_resultsTextArea.setWrapStyleWord(true);
		
		panel.setPreferredSize(new Dimension(500, 150)); // <+++++++++++++++++++++++ KEY SIZING STEP!
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////

	private JPanel initialiseParameterFieldsComponent() {
		
		JPanel panel = new JPanel(new BorderLayout());
		
		JPanel westPanel = new JPanel(new BorderLayout());
		panel.add(westPanel, BorderLayout.WEST);
		westPanel.setBorder(
			new CompoundBorder(
				new EmptyBorder(5, 5, 5, 5),
				new TitledBorder("Oligonucleotide parameters")
				)
			);
		
		JPanel westCenterPanel = new JPanel(new GridLayout(2, 1, 5, 5));
		westPanel.add(westCenterPanel, BorderLayout.CENTER);
		
		JLabel label1 = new JLabel("Required length: ");
		//JLabel label2 = new JLabel("Maximum length: ");
		JLabel label3 = new JLabel("Maximum degenerate bases: ");
		westCenterPanel.add(label1);
		//westCenterPanel.add(label2);
		westCenterPanel.add(label3);
		
		JPanel westEastPanel = new JPanel(new GridLayout(2, 1, 5, 5));
		westPanel.add(westEastPanel, BorderLayout.EAST);
		
		_oligoLengthSpinner = new JSpinner(
			new SpinnerNumberModel(	
				_initialOligoLength,		// Value
				MINIMUM_OLIGO_LENGTH,		// Min 
				MAXIMUM_OLIGO_LENGTH,		// Max
				1							// step
				)
			);
		
		westEastPanel.add(_oligoLengthSpinner);
		((JSpinner.DefaultEditor)_oligoLengthSpinner.
			getEditor()).
				getTextField().
					setColumns(3);
	
		_degenerateBaseSpinner = new JSpinner(
			new SpinnerNumberModel(	
				_initialDegeneracy,						// Value
				0,										// Min
				MAXIMUM_NUMBER_OF_DEGENERATE_BASES,		// Max
				1										// step
				)
			);
	
		westEastPanel.add(_degenerateBaseSpinner);
		((JSpinner.DefaultEditor)_degenerateBaseSpinner.
			getEditor()).
				getTextField().
					setColumns(3);
		
		
		JPanel centerPanel = new JPanel(new GridLayout(1, 1));
		panel.add(centerPanel, BorderLayout.CENTER);
		centerPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
		
		_findAction = new FindAction();
		_findButton = 
			new KeaButton(
				_findAction, 
				KeaButtonColour.Green
				);
		centerPanel.add(_findButton);
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseButtonComponent() {
		
		_resetAction = new ResetAction();
		_cancelAction = new CancelAction();
		_acceptAction = new AcceptAction();
		_acceptAction.setEnabled(false);
		
		JPanel panel = new JPanel(new GridLayout(1, 4, 5, 0));
		panel.setBorder(new EmptyBorder(5, 5, 0, 5));
		
		JButton resetButton = new JButton(_resetAction);
		JButton acceptButton = new JButton(_acceptAction);
		JButton cancelButton = new JButton(_cancelAction);
		
		resetButton.setPreferredSize(new Dimension(150, 30));
		
		panel.add(resetButton);
		panel.add(acceptButton);
		panel.add(cancelButton);
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseProgressBarComponent() {
		
		JPanel panel = new JPanel(new GridLayout(1, 1));
		panel.setBorder(new EmptyBorder(5, 2, 2, 2));
		
		_progressComponent = new ProgressComponent(false);
		panel.add(_progressComponent);
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void findOligos() {
	
		Thread t = new Thread(new Runnable() {public void run(){
			_abortParser = false;
			
			_progressComponent.resetTimer();
			
			// First, extract target sequences from files and store in an array.
			// Second, generate consensus sequence from sequence array.
			Consensus consensus = extractSequencesAndGenerateConsensus();

			// Finally, splice consensus sequence into all possible 
			// oligonucleotides.
			generateOligos(consensus.getStringWithoutSharedGaps());
			
			// All complete - enable all buttons.
			SwingUtilities.invokeLater(new Runnable(){public void run(){
				_cancelAction.setEnabled(true);
				_resetAction.setEnabled(true);
				_acceptAction.setEnabled(true);
				_findAction.setEnabled(true);
				_findButton.setColour(KeaButtonColour.Yellow);
				_oligoLengthSpinner.setEnabled(true);
				_degenerateBaseSpinner.setEnabled(true);
				}});
			
			// note spinner values now (avoid problems if user tries to alter
			// values after running method but before leaving dialog).
			_oligoLengthUsedToGenerateOligos = 
				((Integer)_oligoLengthSpinner.getValue()).intValue();
			_degeneracyUsedToGenerateOligos = 
				((Integer)_degenerateBaseSpinner.getValue()).intValue();	
			
			}}); // End of thread inner class.
		
		// Start thread.	
		t.setPriority(t.MIN_PRIORITY);
		t.start();
		
		
		// Whilst thread is running, disable the following:
		_cancelAction.setEnabled(false);
		_resetAction.setEnabled(false);
		_acceptAction.setEnabled(false);
		_findAction.setEnabled(false);
		
		_oligoLengthSpinner.setEnabled(false);
		_degenerateBaseSpinner.setEnabled(false);
		
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void generateOligos(final String consensus)	{
		
		// Array for storing substrings.
		ArrayList buffer = new ArrayList();
		
		// Initialise array of oligorecord databases objects - each oligo 
		// database stores oligos with different numbers of degenerate bases,
		// with oligo database index number corresponding to degenerate number.
		
		int maxDegeneracy = 
			((Integer)_degenerateBaseSpinner.getValue()).intValue();
		
		OligoDatabase[] oligoDatabases = new OligoDatabase[maxDegeneracy];
		
		// Generate all possible substrings from consensus, given requested
		// oligo length.
		
		int oligoLength = 
			((Integer)_oligoLengthSpinner.getValue()).intValue();
		
		// Move window (of size oligoLength), one base at a time along the
		// consensus sequence.
		for(int i = 0; i <= consensus.length() - oligoLength; i++) {
			
			String oligo = consensus.substring(i, i+oligoLength);
			
			if(noGaps(oligo)) buffer.add(oligo);
			}
		
		String[] oligos = new String[buffer.size()];
		buffer.toArray(oligos);
		
		// Sort oligos according to degeneracy and store in simple oligo objects.
		_oligoDatabase = storeOligos(oligos);
		
		// Lastly, display summary in results text area.
		displayResultsSummary(maxDegeneracy);
		
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void displayResultsSummary(int maxDegeneracy) {
		
		final StringBuffer message = new StringBuffer();
		
		message.append(
			String.valueOf(_oligoDatabase.getTotalNumberOfOligos()) + 
			" potentially useful oligonucleotides have been identified.\n" + 
			"Of these:\n"
			);
		
		OligoCollection collection = null;
		for (int i = 0; i < maxDegeneracy; i++) {
			
			collection = _oligoDatabase.getOligoCollection(i);
			
			message.append(
				String.valueOf(collection.getSize()) + 
				" have " + String.valueOf(i) + 
				" degenerate bases, \n"
				);
			}
		collection = _oligoDatabase.getOligoCollection(maxDegeneracy);
		message.append(
				"and " + String.valueOf(collection.getSize()) + 
				" have " + String.valueOf(maxDegeneracy) + 
				" degenerate bases."
				);
		
		setResultTextAreaText(message.toString());
		
		
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private PrimroseOligoDatabase storeOligos(final String[] oligos) {
		
		// Oligos to be sorted into different oligo collections according to 
		// number of degenerate bases.
		final int maxDegeneracy = 
			((Integer)_degenerateBaseSpinner.getValue()).intValue();
		
		// Create database object and add one oligo collection per degeneracy
		// (index number will correspond to degeneracy number).
		PrimroseOligoDatabase database = new PrimroseOligoDatabase();
		for (int i = 0; i < maxDegeneracy + 1; i++) {
			database.addOligoCollection(
				OligoCollectionFactory.createOligoCollection()
				);
			}
		
		try {
			// Sort oligos into appropriate databases.
			for (int i = 0; i < oligos.length; i++) {

				int degeneracy = SequenceUtil.countDegenerateBases(oligos[i]);

				if (degeneracy <= maxDegeneracy) {
					database.getOligoCollection(degeneracy).addOligo(
						//OligoFactory.createOligo(oligos[i], Orientation.SENSE) // <=====
						new PrimroseOligo(oligos[i], Orientation.SENSE)
						);
					}
				}
			}
		catch(NonIUBCharacterException nice) {
			JOptionPane.showMessageDialog(
				FindOligosDialog.this,
				"Illegal base character encountered whilst generating " + 
				"oligonucleotides.",
				"Non IUB Character Exception",
				JOptionPane.ERROR_MESSAGE
				);
			}
		catch(IllegalSequenceException ise) {
			JOptionPane.showMessageDialog(
				FindOligosDialog.this,
				"Sequence encountered containing both T and U characters" + 
				"whilst generating oligonucleotides.",
				"Illegal Sequence Exception",
				JOptionPane.ERROR_MESSAGE
				);
			}
		catch(NullSequenceException nse) {
			JOptionPane.showMessageDialog(
				FindOligosDialog.this,
				"Null oligonucleotide generated in error.",
				"Null Sequence Exception",
				JOptionPane.ERROR_MESSAGE
				);
			}
		catch(OutOfMemoryError oome) {
			JOptionPane.showMessageDialog(
				FindOligosDialog.this, 
				"<html>" + 
				"The Java virtual machine has run out of memory.<p>" + 
				"Suggestion: try re-running the program from the command<br>" + 
				"line, using the -Xmx option to increase heap memory allocation." + 
				"</html>",
				"Out Of Memory Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		catch(Exception e) {	
			JOptionPane.showMessageDialog(	
				FindOligosDialog.this,
				"Unexpected Exception: " + e.getMessage(),
				"Exception",
				JOptionPane.ERROR_MESSAGE
				);
			e.printStackTrace();
			}

		
		return database;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Returns true if any non-alignment characters are detected in the supplied
	// sequence.
	private boolean noGaps(final String sequence) {
		
		for(int i = 0; i < sequence.length(); i++) {
			if(sequence.charAt(i) == NON_ALIGNMENT_CHARACTER) return false;
			}
		return true;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void setResultTextAreaText(final String text) {
		
		SwingUtilities.invokeLater(new Runnable(){public void run(){
			_resultsTextArea.setText(text);
			}});
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private Consensus extractSequencesAndGenerateConsensus() {
		
		// Stores all consensus strings generated.
		ArrayList consensusBuffer = new ArrayList();
		
		// Output messages stored in this buffer for display in results text box.
		StringBuffer resultsMessage = new StringBuffer();
		
		// Work through each file in database.
		//====================================
		Iterator iterator = _database.getFileIterator();
		while(iterator.hasNext()) {
			
			// Get file.
			DatabaseFile file = (DatabaseFile)iterator.next();
			
			// Get targets chosen to be designed from which come from identified 
			// file.
			Record targets = getTargetsToFindOligosFrom(
				_targetCollection, 
				file
				);
			
			// Do not action if no target records in this file.
			if (!targets.hasChildren()) {continue;}
			
			// Notify user of progress.
			resultsMessage.append(
				"Extracting sequences from " + file.getName() + "...");
			setResultTextAreaText(resultsMessage.toString());
			
			// Parse targets.
			parseAccordingToFileType(file, targets);
			
			// Notify user of progress.
			resultsMessage.append(" Done.\n");
			setResultTextAreaText(resultsMessage.toString());
			
			// Ensure all sequences the same nucleic acid type (note that 
			// RDP database contains mixture of RNA and DNA!)
			String[] sequences = new String[_sequenceBuffer.size()];
			_sequenceBuffer.toArray(sequences); // Convert to string array.
			_sequenceBuffer.clear();  // Empty buffer.
			sequences = 
				SequenceConvertor.convertTo(NucleicAcidType.DNA, sequences);
			
			// Generate consensus.
			// ===================
			// Generate from aligned RDP sequences.
			if (file.getFileType() == FileType.RDP_PHYLO) {
				
				// Notify user of progress.
				resultsMessage.append(
					"Generating consensus sequence from " + file.getName() + 
					" records..."
					);
				setResultTextAreaText(resultsMessage.toString());
				
				// Generate and store consensus object.
				consensusBuffer.add(getConsensusFromAlignedData(sequences));
				
				// Notify user of progress.
				resultsMessage.append(" Done.\n");
				setResultTextAreaText(resultsMessage.toString());
				
				}
			
			// Generate from non-RDP sequences.
			else if (
				file.getFileType() == FileType.FASTA ||
				file.getFileType() == FileType.GENBANK
				) {
				
				// Generate from aligned non-RDP sequences.	
				if (sequencesAreAligned(file, sequences)) {
					
					// Notify user of progress.
					resultsMessage.append(
						"Generating consensus sequence from " + file.getName() + 
						" records..."
						);
					setResultTextAreaText(resultsMessage.toString());

					// Generate and store consensus object.
					consensusBuffer.add(getConsensusFromAlignedData(sequences));
					} 
				// Generate from unaligned non-RDP sequences.	
				else {
					
					// Notify user of progress.
					resultsMessage.append(
						"Generating consensus sequence from " + file.getName() + 
						" records..."
						);
					setResultTextAreaText(resultsMessage.toString());
					
					// Generate and store consensus object.
					consensusBuffer.add(getConsensusFromUnalignedData(sequences));
					}	
				
				// Notify user of progress.
				resultsMessage.append(" Done.\n");
				setResultTextAreaText(resultsMessage.toString());
				
				}
			
			else {
				// Just in case - but shouldn't be thrown as filetype will 
				// already have been checked.
				throw new IllegalArgumentException(
					"Unsupported file type - '" + 
					file.getFileType().toString()+ "'."
					);
				}
			
			} // End of while loop - no more files in database.

		// extraction and consensus sequence generatio complete.  Now, if more 
		// than one consensus (i.e., more than one source file), combine 
		//consensus sequences.
		
		// Scenario one: only one consensus (i.e., targets from single file).
		if (consensusBuffer.size() == 1) {
			return (Consensus)consensusBuffer.get(0);
			}
		
		// Scenario two: more than one consensus - combine.
		else if (consensusBuffer.size() > 1) {
			
			// Notify user of progress.
			resultsMessage.append(
				"Merging consensus sequences..."
				);
			setResultTextAreaText(resultsMessage.toString());
			
			Consensus[] consensusArray = new Consensus[consensusBuffer.size()];
			consensusBuffer.toArray(consensusArray);
			
			// Notify user of progress.
				resultsMessage.append(" Done.\n");
				setResultTextAreaText(resultsMessage.toString());

			// Remove extra long gaps at beginning or end of consensus, which
			// could cause problems when merging consensus sequences. Throw
			// exception if consensus sequences are of widely different lengths.	
			consensusArray = preprocessConsensusArray(consensusArray);	
				
			Consensus consensus = null;	
			try {
				consensus = ConsensusUtil.merge(consensusArray);
				} 
			catch(OutOfMemoryError oome) {
				JOptionPane.showMessageDialog(
					this, 
					"<html>Opps, having problems with your selection. <br>" + 
					"Are sequences of roughly the same length?</html>", 
					"Out Of Memory Error", 
					JOptionPane.ERROR_MESSAGE
					);
				}	
				
			return consensus;
			
			}
		
		// Scenarion three: no consensus sequences!  Shouldn't happen.
		else {
			throw new RuntimeException("No consensus sequences generated!");
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Remove extra long gaps at beginning or end of consensus, which
	// could cause problems when merging consensus sequences. Throw
	// exception if consensus sequences are of widely different lengths.	
	private Consensus[] preprocessConsensusArray(final Consensus[] consensusArray) {
		
		for (int i = 0; i < consensusArray.length; i++) {
			consensusArray[i].trim();
			}
		
		return consensusArray;
		
		} // End of method.	
	
	////////////////////////////////////////////////////////////////////////////
	
	private boolean sequencesAreAligned(
		final DatabaseFile file, 
		final String[] sequences
		) {
		
		// Assume sequences are aligned if all the same length - but check with 
		//user just in case.
		
		final int lengthOfFirstSequence = sequences[0].length(); 
		for(int i = 0; i < sequences.length; i++) {
			if(lengthOfFirstSequence != sequences[i].length()) {
				return false;
				}
			}
		
		if (
			JOptionPane.NO_OPTION == 
				JOptionPane.showConfirmDialog(
					this, 
					"'" + file.getName() + 
					"' appears to be aligned.  Is this correct?", 
					"Just checking...", 
					JOptionPane.YES_NO_OPTION
					)
				) {
			return false;
			} 
		
		return true;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private Consensus getConsensusFromAlignedData(final String[] sequences) {
		
		ConsensusGenerator cg = null;
		
		try {
			cg = ConsensusGeneratorFactory.getConsensusGenerator(
				MultipleSequencesType.Aligned
				);

			cg.addProgressListener(_progressComponent);
			
			return cg.getConsensus(sequences);
			
			} 
		catch(NonIUBCharacterException nice) {

			JOptionPane.showMessageDialog(
				this, 
				"<html>Invalid base characters have been " + 
				"encountered in one or more sequences.<br>" + 
				"Check database files to make sure only valid IUB base" + 
				"characters are used.</html>", 
				"Non IUB Character Exception", 
				JOptionPane.ERROR_MESSAGE
				);

			}
		catch(IllegalSequenceException ise) {

			JOptionPane.showMessageDialog(
				this, 
				"<html>Cannot proceed.  A sequence has been found with a " + 
				"mixture of T and U bases!<br>" + 
				"Please check your file(s) and try again.</html>", 
				"Illegal Sequence Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			} 
		catch(AlignmentException ale) {

			JOptionPane.showMessageDialog(
				this, 
				ale.getMessage(), 
				"Alignment Exception", 
				JOptionPane.ERROR_MESSAGE
				);

			}	
		catch(SequenceTooLongException stle) {

			JOptionPane.showMessageDialog(
				this, 
				stle.getMessage(), 
				"Sequence Too Long Exception", 
				JOptionPane.ERROR_MESSAGE
				);

			}

		catch(Exception e) {
			System.out.println(">>>Unexpected exception!");
			// Print stack trace as this is not supposed to happen!
			e.printStackTrace();
			} 
		
		throw new RuntimeException(">>>Problem with ConsensusGenerator!");
		
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private Consensus getConsensusFromUnalignedData(final String[] sequences) {
		
		ConsensusGenerator cg = null;
		
		try {
			cg = ConsensusGeneratorFactory.getConsensusGenerator(
				MultipleSequencesType.Unaligned
				);
			
			cg.addProgressListener(_progressComponent);

			return cg.getConsensus(sequences);
			
			} 
		catch(NonIUBCharacterException nice) {

			JOptionPane.showMessageDialog(
				this, 
				"<html>Invalid base characters have been " + 
				"encountered in one or more sequences.<br>" + 
				"Check database files to make sure only valid IUB base" + 
				"characters are used.</html>", 
				"Non IUB Character Exception", 
				JOptionPane.ERROR_MESSAGE
				);

			}
		catch(IllegalSequenceException ise) {

			JOptionPane.showMessageDialog(
				this, 
				"<html>Cannot proceed.  A sequence has been found with a " + 
				"mixture of T and U bases!<br>" + 
				"Please check your file(s) and try again.</html>", 
				"Illegal Sequence Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			} 
		catch(AlignmentException ale) {

			JOptionPane.showMessageDialog(
				this, 
				ale.getMessage(), 
				"Alignment Exception", 
				JOptionPane.ERROR_MESSAGE
				);

			}	
		catch(SequenceTooLongException stle) {

			JOptionPane.showMessageDialog(
				this, 
				stle.getMessage(), 
				"Sequence Too Long Exception", 
				JOptionPane.ERROR_MESSAGE
				);

			}

		catch(Exception e) {
			
			JOptionPane.showMessageDialog(
				this, 
				e.getMessage(), 
				">Unexpected Exception!", 
				JOptionPane.ERROR_MESSAGE
				);
			
			// Print stack trace as this is not supposed to happen!
			e.printStackTrace();
			} 
		
		throw new RuntimeException(">>>Problem with ConsensusGenerator!");
		
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Identifies file selected and if ok, selects appropriate parsing code.
	private void parseAccordingToFileType(
		final DatabaseFile file, 
		final Record targets
		) {

		
		try {
		
			if (file.getFileType() == FileType.RDP_PHYLO) {
				parseGenBank(file, targets);
				}
		
			else if (file.getFileType() == FileType.GENBANK) {
				parseGenBank(file, targets);
				}
		
			else if (file.getFileType() == FileType.FASTA) {
				parseFastaFile(file, targets);
				}
		
			}
		
		catch(RecordConflictException rce) {
			JOptionPane.showMessageDialog(
				this, 
				rce.getMessage(), 
				"Record Conflict Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(FileNotFoundException fnfe) {
			JOptionPane.showMessageDialog(
				this, 
				fnfe.toString(),
				"File Not Found Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(IOException ioe) {
			JOptionPane.showMessageDialog(
				this, 
				ioe.toString(),
				"IOException Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(DatabaseParserException dpe) {
			JOptionPane.showMessageDialog(
				this, 
				dpe.toString(),
				"Database Parser Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(OutOfMemoryError oome) {
			JOptionPane.showMessageDialog(
				this, 
				"<html>" + 
				"The Java virtual machine has run out of memory.<p>" + 
				"Suggestion: try altering heap memory allocation with <br>" + 
				"java -Xmx command." + 
				"</html>",
				"Out Of Memory Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(Exception e) {
			JOptionPane.showMessageDialog(
				this, 
				e.toString(),
				"Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			e.printStackTrace();
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void parseFastaFile(final DatabaseFile file, final Record targets) 
		throws
			IOException,
			FileNotFoundException,
            DatabaseParserException,
			RecordConflictException,
			NonIUBCharacterException
		{
				
		// Initialise handler.
		MyFastaFileHandler handler = new MyFastaFileHandler();
		
		// Create parser object.
		FastaFileParserFactory f = FastaFileParserFactory.newInstance();
		FastaSelectiveParser p = f.newSelectiveParser();
		
		// Register a Progress Listener to updte progress bar.
		p.addProgressListener(_progressComponent);
		
		// Register an AbortListener to abort parser if necessary.
		p.addAbortListener(new AbortListener() {
			public boolean abort() {return _abortParser;}
			});
		
		
		// Construct a list of accession numbers to pass to parser.
		ArrayList parseList = new ArrayList();
		Record[] records = targets.getChildrenAsRecords();
		for (int i = 0; i < records.length; i++) {
			parseList.add(records[i].getCode());
			}
		
		// Parse data.	
		p.parse(file, handler, parseList);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void parseGenBank(final DatabaseFile file, final Record targets) 
		throws
			IOException,
            FileNotFoundException,
            ParseException,
			DatabaseParserException,
			RecordConflictException,
			NonIUBCharacterException
			{
		
		// If record source is an RDP .phylo file, get the correct GenBank file.
		DatabaseFile seqFile;
		if (FileType.RDP_PHYLO.isFileType(file)) {
			seqFile = getCorrectRDPFile(file);
			} 
		else {
			seqFile = file;
			}	
		
		// Initialise handler.
		MyGenBankFileHandler handler = new MyGenBankFileHandler();
		
		// Create parser object. 
		GenBankFileParserFactory f = GenBankFileParserFactory.newInstance();
		GenBankSelectiveParser p = f.newSelectiveParser(); 
		
		// Register a Progress Listener to update progress bar.
		p.addProgressListener(_progressComponent);
		
		// Register an AbortListener to abort parser if necessary. 
		p.addAbortListener(new AbortListener() {
			public boolean abort() {return _abortParser;}
			});
		
		// Construct a list of accession numbers to pass to parser.
		ArrayList parseList = new ArrayList();
		Record[] records = targets.getChildrenAsRecords();
		for (int i = 0; i < records.length; i++) {
			parseList.add(records[i].getCode());
			}	
			
		// Parse data.
		p.parse(seqFile, handler, parseList);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Gets RDP .gb file.  If can't find it, asks user for location.
	private DatabaseFile getCorrectRDPFile(final DatabaseFile file) {
		
		// Guess correct path.
		String filepath = file.getAbsolutePath().replaceAll("\\.phylo", ".gb");
		
		// Create DatabaseFile object.
		DatabaseFile dFile = new DatabaseFile(filepath);
		
		// Return DatabaseFile object if file exists.
		if (dFile.isFile()) {return dFile;}
		
		// If does not exist, get user to find it.
		if (
			JOptionPane.showConfirmDialog(
				this, 
				"<html>Can't find " + 
				filepath + ".<br>" +
				"Do you want to find <b>" + dFile.getName() + "</b> yourself?", 
				"File error", 
				JOptionPane.YES_NO_OPTION, 
				JOptionPane.ERROR_MESSAGE
				)
				== JOptionPane.OK_OPTION
			) {
			// Initialise dialog.	
			DatabaseFileChooser fc = new DatabaseFileChooser();
			
			// Remove AcceptAll file filter.
			fc.removeChoosableFileFilter(fc.getAcceptAllFileFilter());
			
			// Add database file filter.
			fc.addChoosableFileFilter(new GenBankFileFilter());
		
			// Set title.
			fc.setDialogTitle("Open database file");
			
			try {
				
				// Open file selected by user.
				if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
					return fc.getSelectedDatabaseFile();
					} 
				
				} 
			
			catch(FileTypeException fte) {
				
				JOptionPane.showMessageDialog(
					this, 
					"Cannot open file - has an unrecognised suffix", 
					"File suffix error", 
					JOptionPane.ERROR_MESSAGE
					);
				}
			}

		return null;
		
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	private class FindAction extends AbstractAction {
		
		public FindAction() {
			putValue(Action.NAME, "Find Oligonucleotides");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Find oligonucleotides shared by all target records."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_I));
			putValue(Action.SMALL_ICON, KeaIcon.InvestigateSequence.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent ae) {	
			
			// First check that a suitable number of targets have been selected.
			int numberOfSelectedTargets = _targetCollection.getNumberSelected();
			
			if (numberOfSelectedTargets == 0) {
				JOptionPane.showMessageDialog(
					FindOligosDialog.this, 
					"<html>No target records have been selected from which to " + 
					"design oligonucleotides.<p>" + 
					"Please select at least one of the target records listed." + 
					"</html>", 
					"No records selected!", 
					JOptionPane.ERROR_MESSAGE
					);
				return;
				}
			
			else if (numberOfSelectedTargets > MAX_NUMBER_OF_TARGETS) {
				JOptionPane.showMessageDialog(
					FindOligosDialog.this, 
					"<html>You have selected far too many records (" + 
					numberOfSelectedTargets + 
					")!  Please revise your selection.<br><br>" + 
					"<b>Recommendation:</b> Best results are obtained when a " + 
					"relatively small number <br>(e.g. no more than " + 
					PREFERRED_MAX_NUMBER_OF_TARGETS + 
					") of well-chosen records are selected.</html>", 
					"Too many targets selected!", 
					JOptionPane.ERROR_MESSAGE
					);
				return;
				}
			
			else if (numberOfSelectedTargets > PREFERRED_MAX_NUMBER_OF_TARGETS) {
				if (
					JOptionPane.NO_OPTION == 
					JOptionPane.showConfirmDialog(
						FindOligosDialog.this, 
						"<html>You have choosen " + numberOfSelectedTargets + 
						" targets. <br><br>" +
						"This is ok, but better results may be achieved by<br>" + 
						"choosing a smaller number (" + 
						PREFERRED_MAX_NUMBER_OF_TARGETS + 
						") of well selected records.<br><br>" + 
						"Do you want to proceed?</html>", 
						"Please note...",
						JOptionPane.YES_NO_OPTION,
						JOptionPane.QUESTION_MESSAGE
						)
					) {
					return;
					}
				}
			findOligos();
			
			}
		
		} // End of inner class.
	
	
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class ResetAction extends AbstractAction {
		
		public ResetAction() {
			putValue(Action.NAME, "Reset");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Resets paramters to default settings."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_R));
			putValue(Action.SMALL_ICON, KeaIcon.Reset.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			
			// Disable reset button.
			_acceptAction.setEnabled(false);
			
			// Empty result text box.
			_resultsTextArea.setText("");
			
			// Re-enable spinners and set with default values.
			//_oligoLengthSpinner.setEnabled(true);
			//_degenerateBaseSpinner.setEnabled(true);
			_oligoLengthSpinner.setValue(new Integer(DEFAULT_OLIGO_LENGTH));
			_degenerateBaseSpinner.setValue(
				new Integer(DEFAULT_NUMBER_OF_DEGENERATE_BASES)
				);
			_findButton.setColour(KeaButtonColour.Green);
			
			deselectAllTargets();
			}
		
		private void deselectAllTargets() {
			for(int i = 0; i < _targetCollection.getSize(); i++) {
				Target t = _targetCollection.getTarget(i);
				if (t.isSelected()) {t.toggleSelection();}
				}
			_model.fireTableDataChanged();
			} 
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class AcceptAction extends AbstractAction {
		
		public AcceptAction() {
			putValue(Action.NAME, "Accept");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Accept oligonucleotides and proceed to next stage."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_A));
			putValue(Action.SMALL_ICON, KeaIcon.Ok.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			
			// Check oligos generated.
			if (insufficientOligos()) {
				JOptionPane.showMessageDialog(
					FindOligosDialog.this, 
					"<html>Your search has generated no oligonucleotides.<br>" + 
					"Try again, or go back and select different target " + 
					"records.</html>", 
					"No oligonucleotides found!", 
					JOptionPane.ERROR_MESSAGE
					);
				return;
				}
			
			_dialogResult = DialogResult.Ok;
			
			// Store user selection for later display.
			_oligoDesignInfo = new OligoDesignInfo(
				_degeneracyUsedToGenerateOligos, 
				_oligoLengthUsedToGenerateOligos, 
				//((Integer)_degenerateBaseSpinner.getValue()).intValue(), 
				//((Integer)_oligoLengthSpinner.getValue()).intValue(),
				_oligoDatabase.getTotalNumberOfOligos()
				);
			
			setVisible(false);
			dispose();
			
			}
		
		private boolean insufficientOligos() {
			if (_oligoDatabase.getTotalNumberOfOligos() == 0) return true;
			return false;
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class CancelAction extends AbstractAction {
		
		public CancelAction() {
			putValue(Action.NAME, "Cancel");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Exit window without accepting oligonucleotides."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_C));
			putValue(Action.SMALL_ICON, KeaIcon.Cancel.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			_dialogResult = DialogResult.Cancel;
			setVisible(false);
			dispose();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class MyFastaFileHandler implements FastaFileHandler {
		
		public void sequence(String sequence) throws DatabaseParserException {
			// Store in temp array until consensus generated.
			_sequenceBuffer.add(sequence);
			}
		
		// NOT IMPLEMENTED.
		public void header(String accession, String text) throws DatabaseParserException, RecordConflictException {}		
		public void recordEnd() throws DatabaseParserException {}		
		public void recordStart() throws DatabaseParserException {}
		
		} // end of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class MyGenBankFileHandler implements GenBankFileHandler {

		public void origin(String sequence) throws DatabaseParserException {
			// Store in temp array until consensus generated.
			_sequenceBuffer.add(sequence);
			}
		
		// NOT IMPLEMENTED.
		public void locus(String locusName, int sequenceLength, MoleculeType moleculeType, GenBankDivision genBankDivision, Date date) throws DatabaseParserException, RecordConflictException {}
		public void accession(String accession) throws DatabaseParserException, RecordConflictException {}
		public void authors(String[] authors) throws DatabaseParserException {}
		public void baseCount(int aCount, int cCount, int gCount, int uCount, int tCount, int othersCount) throws DatabaseParserException {}
		public void comment(String text) throws DatabaseParserException {}
		public void definition(String text) throws DatabaseParserException {}
		public void features(String text) throws DatabaseParserException {}
		public void featuresCDS(String text) throws DatabaseParserException {}
		public void featuresGene(String text) throws DatabaseParserException {}
		public void featuresSource(String text) throws DatabaseParserException {}
		public void journal(String text) throws DatabaseParserException {}
		public void keywords(String text) throws DatabaseParserException {}
		public void medline(String text) throws DatabaseParserException {}
		public void organism(String text) throws DatabaseParserException {}
		public void reference(String text) throws DatabaseParserException {}
		public void source(String text) throws DatabaseParserException {}
		public void title(String text) throws DatabaseParserException {}
		public void version(String version) throws DatabaseParserException {}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	private class MyListSelectionListener implements ListSelectionListener {
		
		public void valueChanged(ListSelectionEvent e) {
			
			//Ignore extra messages.
			if (e.getValueIsAdjusting()) return;
			
			ListSelectionModel lsm = (ListSelectionModel)e.getSource();
			
			if (lsm.isSelectionEmpty()) {
				//no rows are selected - do nothing.
				} 
			else {
				//selectedRow is selected - toggle selection status of 
				// corresponding target.
				int selectedRow = lsm.getMinSelectionIndex();
				
				Target t = _targetCollection.getTarget(selectedRow);
				t.toggleSelection();
				
				// Update table.
				_model.fireTableDataChanged();
				}
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
