// MainClass.java created on 06 March 2004 at 13:48

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.UnsupportedOperationException;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.border.*;
import java.io.*;

import javax.help.*;
import java.net.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import org.xml.sax.*;

import kea.programs.*;
import kea.programs.splashScreens.*;
import kea.programs.dialogBoxes.*;
import kea.programs.dialogBoxes.databaseDialog.*;
import kea.util.*;
import kea.programs.icons.KeaIcon;
import kea.programs.buttons.*;
import kea.database.*;
import kea.graphics.Colour;
import kea.legal.Copyright;
import kea.xml.*;
import kea.system.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Main class for JPrimrose program.
 @author  Kevin Ashelford
 @version 2.17
 */
public class MainClass extends JFrame {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	private static final String HELPSET_LOCATION = 
		MainClass.class.getResource("help/HelpSet.hs").toString();
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private SplashScreen _splashScreen;
	private boolean _showToolTips;
	
	private Database _database; // Stores database records.
	private PrimroseOligoDatabase _oligoDatabase; // Stores all oligos.
	private TargetCollection _targetCollection; // Stores target list.
	private OligoDesignInfo _oligoDesignInfo; // Design info specified by user.
	
	private File _projectFile;
	private ProjectFileFilter _projectFileFilter;
	
	private Action _helpAction;
	private Action _aboutAction;
	private Action _selectDatabaseAction;
	private Action _chooseTargetAction;
	private Action _findOligosAction;
	private Action _searchDatabaseAction;
	private Action _newProjectAction;
	private Action _openProjectAction;
	private Action _saveProjectAction;
	private Action _saveAsProjectAction;
	private Action _notepadAction;
	private Action _exitAction;
	
	private KeaButton _selectDatabaseButton;
	private KeaButton _chooseTargetButton;
	private KeaButton _findOligosButton;
	private KeaButton _searchDatabaseButton;
	
	private HelpSet _helpSet;
	private HelpBroker _helpBroker;
	private ActionListener _helpActionListener;
	
	//##########################################################################
	
	// ENTRY POINT
	
	/**
	 Entry point for program.
	 @param args the command line arguments.
	 */
	public static void main(String[] args) {
		
		LookAndFeelManager.setLookAndFeel();
		
		// Cater for Mac OS X
		if (OperatingSystem.getName() == OperatingSystemName.MaxOSX) {
			System.setProperty("apple.laf.useScreenMenuBar", "true");
			System.setProperty(
				"com.apple.mrj.application.apple.menu.about.name", 
				Info.PROGRAM_NAME
				);
			}
		
		new MainClass();
		
		} // End of main.
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of MainClass.
	 */
	public MainClass() {
		
		// Invoke splash screen.
		_splashScreen = new SimpleSplashScreen(
			Info.SPLASH_SCREEN_DIMENSION, 
			Info.PROGRAM_NAME, 
			Info.PROGRAM_VERSION,
			KeaIcon.LargeProgramIcon.getImageIcon()
			);
		_splashScreen.showSplashScreen(Info.getSplashScreenDuration());
		
		_projectFileFilter = new ProjectFileFilter();
		_showToolTips = Info.DEFAULT_SHOW_TOOLTIPS;
		
		_oligoDatabase = null;
		
		_helpAction = new HelpAction();
		_aboutAction = new AboutAction();
		_exitAction = new ExitAction();
		_newProjectAction = new NewProjectAction(); 
		_newProjectAction.setEnabled(false); // Comment out when fully implemented.*
		_openProjectAction = new OpenProjectAction();
		_openProjectAction.setEnabled(false); // Comment out when fully implemented.*
		_saveProjectAction = new SaveProjectAction();
		_saveAsProjectAction = new SaveAsProjectAction();
		_notepadAction = new NotepadAction(); 
		_notepadAction.setEnabled(false); // Comment out when fully implemented.*
		_selectDatabaseAction = new SelectDatabaseAction();
		_chooseTargetAction = new ChooseTargetAction();
		_findOligosAction = new FindOligosAction();
		_searchDatabaseAction = new SearchDatabaseAction();
		
		/*
		 
		 *New project / notepad work ok, but open/save project facility only
		 partially implemented, therefore all disabled for now.
		 
		 */
		
		
		_selectDatabaseButton = new KeaButton(
			_selectDatabaseAction, 
			KeaButtonColour.Green
			);
		
		_chooseTargetButton = new KeaButton(
			_chooseTargetAction, 
			KeaButtonColour.Green
			);
		
		_findOligosButton = new KeaButton(
			_findOligosAction, 
			KeaButtonColour.Green
			);
		
		_searchDatabaseButton = new KeaButton(
			_searchDatabaseAction, 
			KeaButtonColour.Green
			);
		
		//----------------------------------------------------------------
		// URL url = MainClass.class.getResource("PrimroseHelp/HelpSet.hs");
		
		URL url = null;
		
		try {
			url = new URL((new File(".")).toURL(), HELPSET_LOCATION);
			} 
		catch(MalformedURLException mue) {
			System.out.println(mue.getMessage());
			}
		
		
		try {
			_helpSet = new HelpSet(null, url);
			
			_helpBroker = _helpSet.createHelpBroker();
			_helpActionListener = new CSH.DisplayHelpFromSource(_helpBroker);
			} 
		catch(HelpSetException hse) {
			JOptionPane.showMessageDialog(
				this,
				"<html>Could not find the help files for this application.<br><br>" +
				"The program will still run correctly, but help will be unavailable.</html>",
				"HelpSet Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		finally{
			resetToDefault();
			initialise();
			setVisible(true);
			}
		
		//----------------------------------------------------------------
		
		
		
		} // End of constructor.
	
	//##########################################################################
	
	// PROPERTY ACCESSOR METHODS
	
	//##########################################################################
	
	// METHODS
	
	private void initialise() {
		
		//
		// Define character of overall frame.
		//
		//======================================================================
		this.setResizable(false);
		this.setIconImage(KeaIcon.SmallProgramIcon.getImage());
		this.showToolTips(_showToolTips);
		
		JPanel mainPanel = (JPanel)this.getContentPane();
		mainPanel.setLayout(new BorderLayout());
		
		//
		// Menus.
		//
		//======================================================================
		
		JMenuBar menuBar = new JMenuBar();
		this.setJMenuBar(menuBar);
		menuBar.setBorder(new EmptyBorder(0,0,0,0));
		
		// File
		JMenu fileMenu = new JMenu("File");
		menuBar.add(fileMenu);
		fileMenu.getAccessibleContext().setAccessibleName("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		fileMenu.add(_newProjectAction);
		fileMenu.add(new JSeparator());
		fileMenu.add(_openProjectAction);
		fileMenu.add(_saveProjectAction);
		fileMenu.add(_saveAsProjectAction);
		fileMenu.add(new JSeparator());
		fileMenu.add(_exitAction);
		
		// Options
		JMenu optionsMenu = new JMenu("Options");
		menuBar.add(optionsMenu);
		optionsMenu.getAccessibleContext().setAccessibleName("Options");
		optionsMenu.setMnemonic(KeyEvent.VK_O);
		optionsMenu.add(_selectDatabaseAction);
		optionsMenu.add(_chooseTargetAction);
		optionsMenu.add(_findOligosAction);
		optionsMenu.add(_searchDatabaseAction);
		optionsMenu.add(new JSeparator());
		optionsMenu.add(_notepadAction);
		optionsMenu.add(new JSeparator());
		optionsMenu.add(LookAndFeelManager.getAppearancesSubMenu(this));
		Action showToolTipsAction = new ShowToolTipsAction();
		JCheckBoxMenuItem showToolTipsMenuItem = 
			new JCheckBoxMenuItem(showToolTipsAction);
		showToolTipsMenuItem.setSelected(Info.DEFAULT_SHOW_TOOLTIPS);
		optionsMenu.add(showToolTipsMenuItem);
		
		// Help
		JMenu helpMenu = new JMenu("Help");
		menuBar.add(helpMenu);
		helpMenu.getAccessibleContext().setAccessibleName("Help");
		helpMenu.setMnemonic(KeyEvent.VK_H);
		
		helpMenu.add(_helpAction);
		helpMenu.add(_aboutAction);
		
		//
		// Tool bar.
		//
		//======================================================================
		JToolBar toolBar = new JToolBar();
		mainPanel.add(toolBar, BorderLayout.NORTH);
		toolBar.setRollover(true);
		
		toolBar.add(_newProjectAction);
		toolBar.addSeparator();
		toolBar.add(_openProjectAction);
		toolBar.add(_saveAsProjectAction);
		toolBar.addSeparator();
		toolBar.add(_notepadAction);
		toolBar.addSeparator();
		toolBar.add(_helpAction);
		toolBar.addSeparator();
		
		//
		// Buttons. 
		//
		//======================================================================
		
		JPanel buttonPanel = new JPanel(new GridLayout(4, 1, 0, 20));
		mainPanel.add(buttonPanel, BorderLayout.CENTER);
		buttonPanel.setBorder(new EmptyBorder(20, 50, 35, 50));
		
		buttonPanel.add(_selectDatabaseButton);
		_selectDatabaseButton.setPreferredSize(new Dimension(270, 40));
		
		buttonPanel.add(_chooseTargetButton);
		
		buttonPanel.add(_findOligosButton);
		
		buttonPanel.add(_searchDatabaseButton);
		
		//
		// And finally ...
		//
		//======================================================================
		
		pack(); 
		GUIUtils.setLocationOnceSized(this);	
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void showToolTips(boolean show) {
		ToolTipManager.sharedInstance().setEnabled(show);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/** 
	 Overrides base class method if the WINDOW_CLOSING window event is fired.
	 @param we Event passed from EventListener.
	 */
	protected void processWindowEvent(WindowEvent we) {
		
		// Action if window closing exit is being fired.
		if (we.getID() == WindowEvent.WINDOW_CLOSING) {
			
			// No need for message if database is empty.
			if (_database.isEmpty()) {System.exit(0);}
			
			// There is data in database therefore check with user.
			int userResponse = 
				JOptionPane.showConfirmDialog(
					this, 
					"All unsaved data will be lost. " + 
					"Do you really want to close the program?",
					"Closing program!",
					JOptionPane.YES_NO_OPTION
					);
			if (userResponse == JOptionPane.YES_OPTION) {
				System.exit(0);
				}
			
			} 
		// Otherwise use base class method.
		else {
			super.processWindowEvent(we);
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void selectDatabase() {
		
		OpenDatabaseDialog odd = new OpenDatabaseDialog(this, _database);
		
		odd.addHelpMenuAction((HelpAction)_helpAction);
		odd.addHelpMenuAction((AboutAction)_aboutAction);
		odd.setTitle("Construct Database Dialog");
		
		if (odd.showDatabaseDialog() == DialogResult.Ok) {
			_chooseTargetAction.setEnabled(true);
			_saveAsProjectAction.setEnabled(true);
			_saveProjectAction.setEnabled(true);
			_selectDatabaseButton.setColour(KeaButtonColour.Yellow);
			}
		else {
			
			resetToDefault();
			
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void chooseTarget() {
				
		SimpleChooseTargetDialog ctd = 
			new SimpleChooseTargetDialog(this, _database, _notepadAction);
		
		ctd.addHelpMenuAction((HelpAction)_helpAction);
		ctd.addHelpMenuAction((AboutAction)_aboutAction);
		ctd.setTitle("Choose target records");
		
		if (ctd.showDialog() == DialogResult.Ok) {
			
			// Enable next button and set appropriate colour.
			_findOligosAction.setEnabled(true);
			_findOligosButton.setColour(KeaButtonColour.Green);
			
			// Change current button colour.
			_chooseTargetButton.setColour(KeaButtonColour.Yellow);
			
			// Disable search button if necessary.
			_searchDatabaseAction.setEnabled(false);
			
			// Empty storage objects if necessary.
			_targetCollection = null;
			_oligoDatabase = null; 
			_oligoDesignInfo = null;
			}

		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void findOligos() {
		
		FindOligosDialog fod = new FindOligosDialog(
			this, 
			_database, 
			_targetCollection,
			_oligoDatabase, 
			_oligoDesignInfo,
			_notepadAction
			);
		
		fod.addHelpMenuAction((HelpAction)_helpAction);
		fod.addHelpMenuAction((AboutAction)_aboutAction);
		
		if (fod.showDialog() == DialogResult.Ok) {
			_searchDatabaseAction.setEnabled(true);
			_findOligosButton.setColour(KeaButtonColour.Yellow);
			_oligoDatabase = fod.getOligoDatabase();
			_targetCollection = fod.getTargetCollection(); 
			_oligoDesignInfo = fod.getOligoDesignInfo();
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void searchDatabase() {
		
		SearchDatabaseDialog sdd = new SearchDatabaseDialog(
			this, 
			_database,
			_oligoDatabase,
			_targetCollection,
			_oligoDesignInfo,
			_notepadAction
			);
		
		sdd.addHelpMenuAction(_helpAction);
		sdd.addHelpMenuAction(_aboutAction);
		
		//sdd.showDialog();
		
		sdd.show();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void resetToDefault() {
		
		// Create new database.
		_database = DatabaseFactory.createDatabase(); 
		
		// Reset title bar.
		_projectFile = null;
		updateTitleBar();
		
		// Reset enabled status of buttons/menu items/tool bar items.
		
		_saveAsProjectAction.setEnabled(false);
		_saveProjectAction.setEnabled(false);
		
		_chooseTargetAction.setEnabled(false);
		_findOligosAction.setEnabled(false);
		_searchDatabaseAction.setEnabled(false);
		_selectDatabaseButton.setColour(KeaButtonColour.Green);
		_chooseTargetButton.setColour(KeaButtonColour.Green);
		_findOligosButton.setColour(KeaButtonColour.Green);

		_targetCollection = null;
		_oligoDatabase = null; 
		_oligoDesignInfo = null;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void newProject() {
		
		// Do nothing if not ok to discard existing project.
		if (!okToTrashProject()) {return;}
		
		resetToDefault();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private boolean okToTrashProject() {
		
		// Return true if database is empty.
		if (_database.isEmpty()) {return true;}
		
		// database contains data - ask user for decision.
		int result = JOptionPane.showConfirmDialog(
			this, 
			"Do you want to save the existing project before continuing?",
			"Question", 
			JOptionPane.YES_NO_CANCEL_OPTION,
			JOptionPane.QUESTION_MESSAGE
			);
		
		// Act on decision.
		switch (result) {
			
			case JOptionPane.YES_OPTION:
				return createSaveProjectDialog("Save As");
				
			case JOptionPane.NO_OPTION:
				return true;
				
			case JOptionPane.CANCEL_OPTION:
				return false;
			
			default:
				throw new UnsupportedOperationException(
					"JOptionPane option not suported."
					);
				
			} // End of switch block.
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private boolean createSaveProjectDialog(String title) {
		
		JFileChooser fc = new ProjectFileChooser();
			
		// Remove AcceptAll file filter.
		fc.removeChoosableFileFilter(fc.getAcceptAllFileFilter());
			
		fc.setFileFilter(_projectFileFilter);
		fc.setSelectedFile(_projectFile);
		fc.setDialogTitle(title);
		if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			
			_projectFile = fc.getSelectedFile();
			saveProject();
			updateTitleBar();
			_saveProjectAction.setEnabled(true);
			return true;
			
			}
		else {return false;}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void saveProject() {
		
		try {
			Copyright copyright = 
				new Copyright(Info.PROGRAM_NAME, Info.AUTHOR_NAME);	

			// Create XML document object.	
			Document doc = XMLDocumentFactory.createDocument(_database, copyright); // <================ MUST CHANGE! 
			
			// Open out stream.
			FileWriter out = new FileWriter(_projectFile);

			// Generate file from Document object.
			Transformer transformer = 
				TransformerFactory.newInstance().newTransformer();
			Source source = new DOMSource(doc);
			Result output = new StreamResult(out);
			transformer.transform(source, output);

			// Close stream.
			out.close();
		
			} 
		catch (ParserConfigurationException pce) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" + 
				"A parser configuration error has occurred which has<br>" + 
				"prevented the project from being saved.</html>", 
				"Parser Configuration Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			pce.printStackTrace();
			}
		catch (IOException ioe) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" + 
				"An I/O exception of some sort has occurred which has<br>" + 
				"prevented the project from being saved.</html>", 
				"IO Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			ioe.printStackTrace();
			}
		catch (TransformerConfigurationException tce) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" + 
				"A XML transformer configuration error has occurred which has<br>" + 
				"prevented the project from being saved.</html>", 
				"Transformer Configuration Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			tce.printStackTrace();
			}
		catch (TransformerException te) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" + 
				"A XML transformer error has occurred which has<br>" + 
				"prevented the project from being saved.</html>", 
				"Transformer Configuration Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			te.printStackTrace();
			}		
			
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void updateTitleBar() {
		
		if (_projectFile == null) {setTitle(Info.PROGRAM_NAME);}
		
		else {
			String filename = FileIO.getNameWithoutSuffix(_projectFile);
			setTitle(Info.PROGRAM_NAME + " - " + filename);
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void createOpenProjectDialog() {
		
		// Only action if ok to trash existing project.
		if (!okToTrashProject()) {return;}
		
		// Create Open dialog.
		JFileChooser fc = new JFileChooser();
		
		// Remove AcceptAll file filter.
		fc.removeChoosableFileFilter(fc.getAcceptAllFileFilter());
		
		fc.setFileFilter(_projectFileFilter);
		fc.setDialogTitle("Open Project File");
		
		// Open file selected by user.
		if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
			openProject(fc.getSelectedFile());
			updateTitleBar();
			_saveProjectAction.setEnabled(true);
			_saveAsProjectAction.setEnabled(true);
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void openProject(File file) {
		
		try {
			_projectFile = file;

			// Create DOM parser factory. 
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

			// Create DOM parser.
			DocumentBuilder parser = factory.newDocumentBuilder();

			// Run parser and return resulting Document object.
			Document document = parser.parse(file);


			// Extract details from Document object.
			_database = DatabaseFactory.createDatabase(document); // <================ MUST CHANGE! 
			
			}
		
		catch(NullPointerException npe) {
			JOptionPane.showMessageDialog(
				this,
				"<html>Sorry, cannot proceed; there's something wrong " + 
				"with this file.<br><br>" + 
				"This error message may occur if the file contains empty " + 
				"fields (perhaps it has been manually edited?).",
				"Null Pointer Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			npe.printStackTrace();
			}
		catch (ParserConfigurationException pce) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" + 
				"A parser configuration error has occurred which has<br>" + 
				"prevented the project from being opened.</html>", 
				"Parser Configuration Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			pce.printStackTrace();
			}
		catch(SAXException se) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" + 
				"An error has occurred which has " + 
				"prevented the project from being opened.<br>" + 
				"This suggests there is a problem with the file itself.</html>", 
				"SAX Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			se.printStackTrace();
			}
		
		catch(IOException ioe) {
			JOptionPane.showMessageDialog(
				this,
				"<html>" +
				"Sorry, can not open '" + file.getName() + "'.<br>" + 
				"There appears to be a problem with the file itself.</html>",
				"IO Exception", 
				JOptionPane.ERROR_MESSAGE
				);
			ioe.printStackTrace();
			}

		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void notepad() {
		
		final SimpleNotepadDialog snd =
			new SimpleNotepadDialog(this, true, _database.getNotes());
			
		snd.setBackground(Colour.Chalk);
		snd.setForeground(Colour.NavyBlue);
		snd.setFont(new Font("Dialog", Font.BOLD, 16));
		//snd.setUndecorated(false);
		
		if (DialogResult.Ok == snd.showDialog()) {
			_database.setNotes(snd.getText());
			}
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	private class ProjectFileChooser extends JFileChooser {
	
		//----------------------------------------------------------------------
		
		// Override method.
		public void approveSelection() {
			File file = getFileWithExtension(super.getSelectedFile());
			File directory = super.getCurrentDirectory();
			if (okToSave(file, directory)) {
				super.approveSelection();
				super.setSelectedFile(file);
				}
			else {return;}
			}
		
		//----------------------------------------------------------------------
		
		private File getFileWithExtension(File in) {

			// Take no action if file extension is already present.
			if (_projectFileFilter.accept(in)) {
				return in;
				}

			// Create new file object with correct extension.
			File out = new File(
				in.getAbsoluteFile() + 
				"." + 
				_projectFileFilter.getExtension()
				);

			return out;

			}
			
		//----------------------------------------------------------------------
		
		private boolean okToSave(File file, File directory) {

			// Get list of files in current directory.
			File[] files = directory.listFiles(_projectFileFilter);

			// Check user supplied name with these.
			for (int i = 0; i < files.length; i++) {

				String name1 = files[i].getName();
				String name2 = file.getName();

				if (name1.equals(name2)) {
					return okToOverwrite(name1);
					}

				}

			// No existing file with same name therefore ok to save.
			return true;

			}
			
		//----------------------------------------------------------------------
		
		private boolean okToOverwrite(String name) {

			if (JOptionPane.showConfirmDialog(
				this,
				"'" + name + "' already exists.  Do you want to replace it?",
				"Warning",
				JOptionPane.YES_NO_OPTION,
				JOptionPane.INFORMATION_MESSAGE
				) == JOptionPane.YES_OPTION) {return true;} else {return false;}

			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Load database button.
	private class SelectDatabaseAction extends AbstractAction {
		public SelectDatabaseAction() {
			putValue(Action.NAME, "Construct Database");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"<html>Load sequence records into the program. </html>"
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_S));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0)
				);
			putValue(Action.SMALL_ICON, KeaIcon.Database.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			selectDatabase();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Identify target group button.
	private class ChooseTargetAction extends AbstractAction {
		public ChooseTargetAction() {
			putValue(Action.NAME, "Specify Target Taxon");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"<html>Identify the target taxon that the oligonucleotides " + 
				"are meant to describe</html>"
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_I));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_F2, 0)
				);
			putValue(Action.SMALL_ICON, KeaIcon.ChooseTarget.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			chooseTarget();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Generate oligos button.
	private class FindOligosAction extends AbstractAction {
		public FindOligosAction() {
			putValue(Action.NAME, "Find Oligonucleotides");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"<html>Identify potentially useful oligonucleotides from the " + 
				"specified target taxon.</html>"
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_G));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_F3, 0)
				);
			putValue(Action.SMALL_ICON, KeaIcon.InvestigateSequence.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			findOligos();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Check oligos against database button.
	private class SearchDatabaseAction extends AbstractAction {
		public SearchDatabaseAction() {
			putValue(Action.NAME, "Check Oligonucleotides Against Database");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"<html>Assesses suitability of identified oligonucleotides.</html>"
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_C));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_F4, 0)
				);
			putValue(Action.SMALL_ICON, KeaIcon.AnalyseDatabase.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			searchDatabase();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Switch off tooltips.
	private class ShowToolTipsAction extends AbstractAction {
		public ShowToolTipsAction() {
			putValue(Action.NAME, "Show tool tips");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Enable or disable these annoying messages."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_S));
			putValue(Action.SMALL_ICON, KeaIcon.ShowToolTips.getImageIcon());
			} 
		public void actionPerformed(ActionEvent ae) {
			_showToolTips = !_showToolTips;
			showToolTips(_showToolTips);
			}
		} // End of inner class.
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Help.
	private class HelpAction extends AbstractAction {	
		public HelpAction() {	
			putValue(Action.NAME, "Help");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Get help on this program."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_H));
			putValue(Action.SMALL_ICON, KeaIcon.Help.getImageIcon());
			}
		public void actionPerformed(ActionEvent ae) {
			
			if (_helpActionListener == null) {
				JOptionPane.showMessageDialog(
					MainClass.this, 
					"<html>Help files not available. Sorry!",
					"HelpSet Error",
					JOptionPane.ERROR_MESSAGE
					);
				}
			else {
				_helpActionListener.actionPerformed(ae);
				}
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// About.
	private class AboutAction extends AbstractAction {
		public AboutAction(){
			putValue(Action.NAME, "About");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Basic information about this program."
				);
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_A));
			putValue(Action.SMALL_ICON, KeaIcon.SmallProgramIcon.getImageIcon());
			} // End of inner class constructor.
		public void actionPerformed(ActionEvent ae) {
			
			AboutDialogBox about = new AboutDialogBox(
				MainClass.this, 
				_splashScreen
				);
			about.show();
			
			} // End of inner class method.
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Exit.
	private class ExitAction extends AbstractAction {
		public ExitAction() {
			putValue(Action.NAME, "Exit");
			putValue(Action.SHORT_DESCRIPTION, "Exit program.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_X));
			putValue(Action.SMALL_ICON, KeaIcon.Exit.getImageIcon());
			} 
		public void actionPerformed(ActionEvent ae) {
			WindowEvent we = new WindowEvent(
				MainClass.this, 
				WindowEvent.WINDOW_CLOSING
				);
			processWindowEvent(we);
			} 
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class NewProjectAction extends AbstractAction {
		
		public NewProjectAction() {
			putValue(Action.NAME, "New Project");
			putValue(Action.SHORT_DESCRIPTION, "Start a new project.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_N));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(Action.SMALL_ICON, KeaIcon.New.getImageIcon());
			} 
		
		public void actionPerformed(ActionEvent e) {
			newProject();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class OpenProjectAction extends AbstractAction {
		
		public OpenProjectAction() {
			putValue(Action.NAME, "Open Project");
			putValue(Action.SHORT_DESCRIPTION, "Open a previously saved project.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_O));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_O, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(Action.SMALL_ICON, KeaIcon.Open.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			createOpenProjectDialog();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class SaveProjectAction extends AbstractAction {
		
		public SaveProjectAction() {
			putValue(Action.NAME, "Save Project");
			putValue(Action.SHORT_DESCRIPTION, "Saves existing project.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_S));
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(Action.SMALL_ICON, KeaIcon.Save.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			// Call save dialog if no filename associated with current data. 
			if (_projectFile == null) {
				createSaveProjectDialog("Save Project File As");
				}
			// Otherwise, save file with existing filename.
			else {
				saveProject();	
				}
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class SaveAsProjectAction extends AbstractAction {
		
		public SaveAsProjectAction() {
			putValue(Action.NAME, "Save As...");
			putValue(Action.SHORT_DESCRIPTION, "Saves project.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_A));
			putValue(Action.SMALL_ICON, KeaIcon.Save.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			createSaveProjectDialog("Save Project File As");
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class NotepadAction extends AbstractAction {
		
		public NotepadAction() {
			putValue(Action.NAME, "Notepad");
			putValue(Action.SHORT_DESCRIPTION, "Store project comments.");
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_F5, 0)
				);
			putValue(Action.SMALL_ICON, KeaIcon.Notepad.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			notepad();
			}
		
		} // End of inner class.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
