// OligoSearchResult.java created on 17 May 2004 at 12:59

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import java.util.*;

import kea.database.*;
import kea.sequences.nucleicAcid.search.*;
import kea.sequences.nucleicAcid.tm.*;
import kea.sequences.nucleicAcid.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Stores outcome of database search with specified oligonucleotide.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public class OligoSearchResult {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private final Oligo _oligo;
	private int _targetHits;
	private int _nonTargetHits;
	
	private boolean _failed;
	
	private final BasicGCContentCalculator _gcCalculator;
	
	private final HashMap _searchResults;  // Key=Record, value=SearchResult.

	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of <code>OligoSearchResult</code>.
	 @param oligo Oligo object.
	 */
	public OligoSearchResult(final Oligo oligo) {
		_oligo = oligo;
		_gcCalculator = new BasicGCContentCalculator(oligo.getSequence());
		
		_searchResults = new HashMap();

		_failed = false;
		} // End of constructor.
	
	//##########################################################################
	
	// METHODS
	
	/**
	 Returns Oligo object.
	 @return Oligo object.
	 */
	public Oligo getOligo() {
		return _oligo;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns SearchResult corresponding to specified Record object.
	 @param record Record object.
	 @return SearchResult.
	 */
	public SearchResult getSearchResult(final Record record) {
		return (SearchResult)_searchResults.get(record);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Stores search result for specified Record object.
	 @param record Record object.
	 @param searchResult SearchResult object.
	 */
	public void addSearchResult(
		final Record record, 
		final SearchResult searchResult
		) {
		
		// First store result.	
		_searchResults.put(record, searchResult);
		
		// Increment appropriate hit counter if successful search.
		if (searchResult != SearchResult.NoMatch) {
			if (record.getRecordType() == RecordType.Target) _targetHits++;
			else _nonTargetHits++;
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the total number of database hits (target and non-target).
	 @return Hits.
	 */
	public int getHits() {
		return _targetHits + _nonTargetHits;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns total number of target hits.
	 @return Target hits.
	 */
	public int getTargetHits() {
		return _targetHits;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns total number of non-target hits.
	 @return Non-target hits.
	 */
	public int getNonTargetHits() {
		return _nonTargetHits;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns melting temperature of oligo, calculated according to equation 
	 indicated by supplied argument.
	 @param tmCalculatorType Melting temperature equation to use.
	 @return Melting temperature.
	 */
	public int getMeltingTemperature(TmCalculatorType tmCalculatorType) {
		throw new UnsupportedOperationException();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the number of degenerate bases within the oligo sequence associated
	 with this object.
	 @return number of degenerate bases.
	 */
	public int getDegeneracy() {
		return SequenceUtil.countDegenerateBases(_oligo.getSequence());
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the percentage GC content of the oligo.
	 @return Percentage GC content.
	 */
	public double getPercentGCContent() {
		return _gcCalculator.getPercentGCContent();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the number of self complementary bases within the oligonucleotide.
	 @return Number of self complementary bases.
	 */
	public int getNumberOfSelfComplementaryBases() {
		throw new UnsupportedOperationException();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Records whether search has failed (exceeded non-target record limit).
	 @param b True if failed.
	 */
	public void setFailed(boolean b) {
		_failed = b;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns true if search has failed (exceeded non-target record limit).
	 @return True if failed.
	 */
	public boolean hasFailed() {
		return _failed;
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
