// PrimroseOligo.java created on 24 May 2004 at 13:26

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import java.util.regex.*;
import java.util.*;

import kea.sequences.nucleicAcid.*;
import kea.exceptions.*;
import kea.database.*;
import kea.sequences.nucleicAcid.search.*;
import kea.sequences.nucleicAcid.tm.*;
import kea.sequences.nucleicAcid.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Implementation of Oligo interface.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public final class PrimroseOligo implements Oligo {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private final String _sequence;
	private final Orientation _orientation;
	private final Pattern _pattern;
	private final Matcher _matcher;
	
	private int _targetHits = 0;
	private int _totalTargets = 0;
	private int _nonTargetHits = 0;
	private int _totalNonTargets = 0;
	private boolean _failed = false;
	private final BasicGCContentCalculator _gcCalculator;
	
	private boolean _searchComplete = false;
	
	private final HashMap _searchOutcomes;  // Key=Record, value=SearchOutcome.
	private final HashMap _sequenceLengths; // Key=Record, value=Integer.
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of <code>PrimroseOligo</code>.
	 */
	public PrimroseOligo(
		final String sequence, 
		final Orientation orientation
		)
		throws 
			NonIUBCharacterException,
			IllegalSequenceException,
			NullSequenceException
		{
		
		if (SequenceUtil.containsNonIUBCharacter(sequence)) {
			throw new NonIUBCharacterException(
				"Oligonucleotide contains one or more illegal base characters."
				);
			}
		if (SequenceUtil.containsBothTAndUBases(sequence)) {
			throw new IllegalSequenceException(
				"Oligonucleotide contains both T and U bases.", sequence
				);
			}
		
		if (sequence.length() == 0 || sequence == null) {
			throw new NullSequenceException(
				"No oligonucleotide supplied."
				);
			}
			
		//_sequence = sequence;
		
		// Convert to DNA (makes comparison with sequences easier).
		_sequence = SequenceConvertor.convertTo(NucleicAcidType.DNA, sequence);
		
		_orientation = orientation;
		_pattern = SequenceConvertor.getSimplePattern(_sequence);
		//_pattern = SequenceConvertor.getPattern(_sequence);
		_matcher = _pattern.matcher("");
		
		_gcCalculator = new BasicGCContentCalculator(_sequence);
		
		_searchOutcomes = new HashMap();
		_sequenceLengths = new HashMap();
		
		} // End of constructor.
	
	//##########################################################################
	
	// METHODS

	public Orientation getOrientation() {
		return _orientation;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	public String getSequence() {
		return _sequence;
		} // End of method.

	////////////////////////////////////////////////////////////////////////////
	
	public Pattern getPattern() {
		return _pattern;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	public Matcher getMatcher() {
		return _matcher;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	public Matcher getMatcher(final CharSequence input) {
		return _matcher.reset(input);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Scores search result (target/non-target) according specified Record object.
	 @param record Record object.
	 @param searchResult SearchResult object.
	 */
	public void incrementScore(
		final Record record, 
		final SearchResult searchResult
		) {
			
		// Determine whether record is a target or not and increment appropriate
		// counter.	
		RecordType type = record.getRecordType();
		if (type == RecordType.Target) _totalTargets++; else _totalNonTargets++;
		
		// Next, increment appropriate hit counter if successful search.
		if (searchResult != SearchResult.NoMatch) {
			
			if (type == RecordType.Target) _targetHits++;
			else _nonTargetHits++;
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns all Record objects encountered by oligo during search.
	 @return Array of Record objects.
	 */
	public Record[] getRecords() {
		
		Set keys = _searchOutcomes.keySet();
		Record[] records = new Record[keys.size()];
		keys.toArray(records);
		return records;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns all target Record objects encountered by oligo during search.
	 @return Array of Record objects.
	 */
	public Record[] getTargetRecords() {
		
		ArrayList buffer = new ArrayList();
		Record[] records = getRecords();
		
		for (int i = 0; i < records.length; i++) {
			if (records[i].getRecordType() == RecordType.Target) {
				buffer.add(records[i]);
				}
			}
		
		Record[] targets = new Record[buffer.size()];
		buffer.toArray(targets);
		return targets;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns all non-target Record objects encountered by oligo during search.
	 @return Array of Record objects.
	 */
	public Record[] getNonTargetRecords() {
		
		ArrayList buffer = new ArrayList();
		Record[] records = getRecords();
		
		for (int i = 0; i < records.length; i++) {
			if (records[i].getRecordType() != RecordType.Target) {
				buffer.add(records[i]);
				}
			}
		
		Record[] nonTargets = new Record[buffer.size()];
		buffer.toArray(nonTargets);
		return nonTargets;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Store Record object and corresponding SearchOutcome.  
	 @param record Record object.
	 @param searchOutcome for Record.
	 */
	public void addSearchOutcome(
		final Record record, 
		final SearchOutcome searchOutcome
		) {
		_searchOutcomes.put(record, searchOutcome);	
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Stores sequence length of record object for display purposes.
	 @param record Record object.
	 @param sequenceLength Sequence length.
	 */
	public void addSequenceLength(
		final Record record,
		final int sequenceLength
		) {
		_sequenceLengths.put(record, new Integer(sequenceLength));
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns sequence length of supplied record object.
	 @param record Record object.
	 @return Sequence length.
	 */
	public int getSequenceLength(final Record record) {
		return ((Integer)_sequenceLengths.get(record)).intValue();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns search result corresponding to Record object supplied.
	 @param record Record object in question.
	 @return Outcome of search between oligo and Record object.
	 */
	public SearchResult getSearchResult(final Record record) {
		SearchOutcome so = (SearchOutcome)_searchOutcomes.get(record);
		return so.getSearchResult();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns start position of oligo within specified Record object.
	 @param record Record object.
	 @return Start position.
	 */
	public int getStartPosition(Record record) {
		SearchOutcome so = (SearchOutcome)_searchOutcomes.get(record);
		return so.getStartPosition();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns end position of oligo within specified Record object.
	 @param record Record object.
	 @return End position.
	 */
	public int getEndPosition(Record record) {
		SearchOutcome so = (SearchOutcome)_searchOutcomes.get(record);
		return so.getEndPosition();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the total number of database hits (target and non-target).
	 @return Hits.
	 */
	public int getHits() {
		return _targetHits + _nonTargetHits;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns total number of target hits.
	 @return Target hits.
	 */
	public int getTargetHits() {
		return _targetHits;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns total number of non-target hits.
	 @return Non-target hits.
	 */
	public int getNonTargetHits() {
		return _nonTargetHits;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns percentage of target records hit.
	 @return Percentage.
	 */
	public float getPercentTargetHits() {
		
		return ((float)_targetHits / (float)_totalTargets) * 100;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns percentage of non-target records hit.
	 @return Percentage.
	 */
	public float getPercentNonTargetHits() {
		
		return ((float)_nonTargetHits/(float)_totalNonTargets) * 100;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns melting temperature of oligo, calculated according to equation 
	 indicated by supplied argument.
	 @param tmCalculatorType Melting temperature equation to use.
	 @return Melting temperature.
	 */
	public double getMeltingTemperature(TmCalculatorType tmCalculatorType) {
		
		if (tmCalculatorType == TmCalculatorType.Wallace) {
			WallaceRuleTmCalculator tm = new WallaceRuleTmCalculator(
				NucleicAcidFactory.createOligonucleotide(_sequence)
				);
			return tm.getResult();
			}
		
		if (tmCalculatorType == TmCalculatorType.ChesterMarshak) {
			ChesterMarshakTmCalculator tm = new ChesterMarshakTmCalculator(
				NucleicAcidFactory.createOligonucleotide(_sequence)
				);
			return tm.getResult();
			}
		
		
		throw new IllegalArgumentException(
			"'" + tmCalculatorType.Wallace.toString() + 
			"' is not currently supported."
			);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the number of degenerate bases within the oligo sequence associated
	 with this object.
	 @return number of degenerate bases.
	 */
	public int getDegeneracy() {
		return SequenceUtil.countDegenerateBases(this.getSequence());
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the percentage GC content of the oligo.
	 @return Percentage GC content.
	 */
	public double getPercentGCContent() {
		return _gcCalculator.getPercentGCContent();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns the number of self complementary bases within the oligonucleotide.
	 @return Number of self complementary bases.
	 */
	public int getNumberOfSelfComplementaryBases() {
		throw new UnsupportedOperationException();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Records whether search has failed (exceeded non-target record limit).
	 @param b True if failed.
	 */
	public void setFailed(boolean b) {
		_failed = b;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns true if search has failed (exceeded non-target record limit).
	 @return True if failed.
	 */
	public boolean hasFailed() {
		return _failed;
		} // End of method.
	
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Resets object - empties all storage.
	 */
	public void reset() {
		_targetHits = 0;
		_nonTargetHits = 0;
		_totalTargets = 0;
		_totalNonTargets = 0;
		_failed = false;
		_sequenceLengths.clear();
		_searchOutcomes.clear();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Sets 'search complete' status for object.
	 @param b True if search has finished and so data available for display.
	 */
	public void setSearchComplete(boolean b) {
		_searchComplete = b;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns true if search completed, and therefore record information 
	 available for display. 
	 @return True if record-related information is available.
	 */
	public boolean searchComplete() {
		return _searchComplete;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	public String toString() {
		return _sequence;
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
