// Report.java created on 17 May 2004 at 16:55

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.border.*;
import javax.swing.table.*;
import java.util.*;
import javax.print.*;
import javax.print.attribute.*;
import javax.print.attribute.standard.*;
import java.io.*;
import java.awt.image.*;
import javax.imageio.*;
import javax.swing.tree.*;
import javax.swing.event.*;

import kea.database.*;
import kea.programs.database.*;
import kea.programs.icons.KeaIcon;
import kea.programs.buttons.*;
import kea.programs.progress.*;
import kea.event.*;
import kea.programs.dialogBoxes.*;
import kea.programs.plot.*;
import kea.sequences.nucleicAcid.search.*;
import kea.sequences.nucleicAcid.*;
import kea.exceptions.*;
import kea.io.*;
import kea.util.*;
import kea.programs.text.*;
import kea.programs.icons.KeaIcon;
import kea.graphics.Colour;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Oligo search report.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public class Report extends JDialog {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	private static final int DEFAULT_MAX_MISMATCHES = 3;
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private int _maxMismatches;
	
	private OligoPanel _oligoPanel;
	private Database _database;
	private PrimroseOligo _oligo;
	
	private JMenu _helpMenu;
	
	private Action _searchAction;
	private Action _abortAction;
	private Action _showDataAction;
	private Action _printGraphAction;
	private Action _saveGraphAction;
	private Action _enableSelectionAction;
	private Action _notepadAction;

	private JSpinner _mismatchSpinner;
	private boolean _abortParser;
	private boolean _dataGenerated;
	private ProgressComponent _progressComponent;
	private DatabaseTree _databaseTree;
	private boolean _dataTableVisible;
	private ReportGraphTableDialog _graphTableDialog;
	private JPopupMenu _graphPopupMenu;
	private Graph _graph;
	private File _graphFile;
	private JTable _table;
	
	// Default cell renderers (prior to search).
	private PrimroseTreeCellRenderer _primroseTreeCellRenderer;
	private PrimroseTableCellRenderer _primroseTableCellRenderer;
	
	// Cell renderers for displaying search results.
	private ReportTreeCellRenderer _reportTreeCellRenderer;
	private ReportTableCellRenderer _reportTableCellRenderer;
	
	// For adding additional sequence files to database:
	OpenDatabaseAction _openDatabaseAction;
	
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of <code>Report</code>.
	 @param owner Owning Frame.
	 @param oligo Oligonucleotide to display.
	 @param database Database object to display.
	 */
	public Report( 
		final Frame owner,
		final PrimroseOligo oligo,
		final Database database
		) {
		super(owner, "Report", true);	
		initialise(oligo, database);
		} // End of constructor.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Creates a new instance of <code>Report</code>.
	 @param owner Owning dialog box.
	 @param oligo Oligonucleotide to display.
	 @param database Database object to display.
	 */
	public Report( 
		final Dialog owner,
		final PrimroseOligo oligo,
		final Database database
		) {
		super(owner, "Report", true);	
		initialise(oligo, database);
		} // End of constructor.
	
	//##########################################################################
		
	// METHODS
	
	private void initialise(PrimroseOligo oligo, Database database) {
		
		_abortParser = true;
		
		_maxMismatches = DEFAULT_MAX_MISMATCHES;
		
		_notepadAction = new NotepadAction();
		
		_primroseTreeCellRenderer = new PrimroseTreeCellRenderer();
		_reportTreeCellRenderer = new ReportTreeCellRenderer();
		_primroseTableCellRenderer = new PrimroseTableCellRenderer();
		_reportTableCellRenderer = new ReportTableCellRenderer();
		
		_database = database;
		
		// Initialise now - needed by OpenDatabaseAction.
		_progressComponent = new ProgressComponent(true); 
		
		// For adding additional sequence files to database:
		_openDatabaseAction = new OpenDatabaseAction(
			this, 
			_database, 
			_progressComponent, 
			false, // Don't clear database when new files opened.
			false, // Don't show mnemonic etc.
			"Add Sequences to Database",
			"Add additional sequence files to existing database.", 
			KeaIcon.OpenDatabaseFile.getImageIcon()
			);
		
		_oligo = oligo;
		_oligoPanel = new OligoPanel(
			oligo.getSequence(), 
			oligo.getOrientation(), 
			"Oligonucleotide (5' - 3')", 
			true
			);
		
		_graphFile = null;
		
		// Graph data initially not visible.
		_dataTableVisible = false;
		
		// Initialise Actions.
		_showDataAction = new ShowDataAction();
		_printGraphAction = new PrintGraphAction();
		_saveGraphAction = new SaveGraphAction();
		_enableSelectionAction = new EnableSelectionAction();
		
		// Disable until search complete.
		_showDataAction.setEnabled(false);
		_printGraphAction.setEnabled(false);
		_saveGraphAction.setEnabled(false);
		_enableSelectionAction.setEnabled(false);
		
		
		// Initialise corresponding data dialog.
		_graphTableDialog = new ReportGraphTableDialog(this, false); 
		_graphTableDialog.addWindowListener(
			new WindowAdapter() {
				public void windowClosing(WindowEvent we){
				// Update tool bar button.
				_dataTableVisible = !_dataTableVisible; 
				}});
		
		//======================================================================
		
		// Menu bar.
		setJMenuBar(menuBarComponent());
		
		//======================================================================
		
		JPanel mainPanel = (JPanel)getContentPane();
		mainPanel.setLayout(new BorderLayout());
		JPanel innerPanel = new JPanel(new BorderLayout());
		
		// mainPanel north - tool bar.
		mainPanel.add(toolBarComponent(), BorderLayout.NORTH);
		
		// mainPanel center - innerPanel.
		mainPanel.add(innerPanel, BorderLayout.CENTER);
		
		// mainPanel south - progress component.
		mainPanel.add(initialiseProgressPanel(), BorderLayout.SOUTH);
		
		
		// innerPanel north - information component.
		innerPanel.add(_oligoPanel, BorderLayout.NORTH);
		
		// innerPanel center - tree/table component.
		innerPanel.add(initialiseDatabasePanel(), BorderLayout.CENTER);
		
		// innerPanel south - button component.
		innerPanel.add(initialiseButtonPanel(), BorderLayout.SOUTH);
		
		//======================================================================
		
		this.setSize(800, 600);
		kea.util.GUIUtils.setLocationOnceSized(this);
		kea.util.GUIUtils.setWindowMinimumSize(this, getSize());
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Displays dialog box.
	 @return DialogResult.
	 */
	public DialogResult showDialog() {
		setVisible(true);
		return DialogResult.Ok;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JToolBar toolBarComponent() {
		
		JToolBar toolBar = new JToolBar();
		toolBar.setRollover(true);
		
		toolBar.add(_openDatabaseAction);	
		toolBar.addSeparator();
		toolBar.add(_oligoPanel.getOpenAction());
		toolBar.add(_oligoPanel.getSaveAsAction());
		toolBar.add(_notepadAction);
		toolBar.addSeparator();
		toolBar.add(_showDataAction);
		toolBar.add(_printGraphAction);
		toolBar.add(_saveGraphAction);
		toolBar.addSeparator();
		toolBar.add(_oligoPanel.getCutAction());
		toolBar.add(_oligoPanel.getCopyAction());
		toolBar.add(_oligoPanel.getPasteAction());
		toolBar.add(_oligoPanel.getDeleteAction());
		toolBar.add(_oligoPanel.getUndoAction());
		toolBar.add(_oligoPanel.getRedoAction());
		toolBar.addSeparator();
		toolBar.add(_enableSelectionAction);
		toolBar.addSeparator();
		
		return toolBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseDatabasePanel() {
		
		JPanel panel = new JPanel(new GridLayout());
		
		// Initialise tree with empty database.
		_databaseTree = new DatabaseTree(_database); 
		//_databaseTree.enableTreeRecordSelection(false);
		_databaseTree.getTree().setCellRenderer(_primroseTreeCellRenderer);
		_databaseTree.getTree().addTreeSelectionListener(new SelectionListener());
		_databaseTree.getTree().addTreeExpansionListener(new ExpansionListener());
		// Add additional mouse listener to update table whenever record is
		// selected / deselected.
		_databaseTree.getTree().addMouseListener(new TreeMouseListener());
		
		JSplitPane secondarySplitPane = new JSplitPane(
			JSplitPane.VERTICAL_SPLIT,
			table(),
			graph()
			);
		secondarySplitPane.setDividerLocation(200);
		secondarySplitPane.setBorder(new EmptyBorder(0, 0, 0, 0));
		secondarySplitPane.setDividerSize(7);
		secondarySplitPane.setOneTouchExpandable(true);
	
		
		JSplitPane primarySplitPane = new JSplitPane(
			JSplitPane.HORIZONTAL_SPLIT,
			_databaseTree, 
			secondarySplitPane
			);
		primarySplitPane.setDividerLocation(300);
		primarySplitPane.setBorder(new EmptyBorder(0, 0, 0, 0));
		primarySplitPane.setDividerSize(7);
		primarySplitPane.setOneTouchExpandable(true);
		panel.add(primarySplitPane);	
		
		panel.setEnabled(false);

		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JComponent table() {
		
		ReportTableModel model = 
			new ReportTableModel(RecordFactory.createDummyRecord(), null);
		
		// Initialise table with TableModel.
		_table = new JTable(model);
		
		// Add TableCellRenderer for dealing with name and icon objects.
		_table.setDefaultRenderer(Record.class, _primroseTableCellRenderer);
		
		// Determine how columns should be resized.
		_table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
		
		// Set column widths.
		model.setColumnWidths(_table);
		
		// Do not display grid.
		_table.setShowGrid(false);
		
		// Set selection colours.
		_table.setSelectionBackground(
			kea.programs.dialogBoxes.databaseDialog.Info.BACKGROUND_SELECTION_COLOUR
			);
		_table.setSelectionForeground(Color.BLACK);
		
		// Add mouse listener for record selection/deselection.
		_table.addMouseListener(new TableMouseListener());
		
		// Add to scroll pane, set background to white and return.
		JScrollPane sp = new JScrollPane(_table);
		sp.getViewport().setBackground(Color.WHITE);
		return sp;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JComponent graph() {
		
		Axis xAxis = AxisFactory.createAxis(
			"Number of mismatches", 
			AxisType.X, 
			0, 
			_maxMismatches, 
			1, 
			0, 
			ScaleType.Linear
			);
		
		Axis yAxis = AxisFactory.createAxis(
			"% records targeted", 
			AxisType.Y, 
			0, 
			100, 
			20, 
			4, 
			ScaleType.Linear
			);
		
		// Initialise graph object.
		_graph = new Graph(xAxis, yAxis);
		
		_graph.setAntiAliasing(true);
		_graph.setTopMargin(30);
		_graph.setBottomMargin(50);
		_graph.setLeftMargin(80);
		_graph.setRightMargin(50);
		_graph.setTickToLabelGap(5);
		_graph.setXAxisLabelToBorderGap(10);
		_graph.setYAxisLabelToBorderGap(25);
		
		// Key.
		_graph.setKeyVisible(true);
		_graph.setKeyBorderVisible(true);
		_graph.setKeyOpaque(true);
		//_graph.setKeyLocation(new Point(100, 2));
		_graph.setKeyMargin(5);
		_graph.setKeyFont(new Font("Dialog", Font.PLAIN, 12));
		_graph.setKeyDefaultLocation(false);
		
		DefaultAxisRenderer r = new DefaultAxisRenderer();
		r.setTickLabelDecimalPlaces(0); 
		r.setMajorGridLinesVisible(true);
		r.setMajorGridLineColor(Color.GREEN); 
		r.setMajorGridLineStyle(LineStyle.Solid); 
		r.setMajorGridLineWidth(1);
		r.setMinorGridLinesVisible(true); 
		r.setMinorGridLineColor(Color.CYAN); 
		r.setMinorGridLineStyle(LineStyle.Solid);
		r.setMinorGridLineWidth(1); 
		_graph.setAxisRenderer(r);
		
		
		_graphPopupMenu = new JPopupMenu();
		_graphPopupMenu.add(_showDataAction);
		_graphPopupMenu.add(_printGraphAction);
		_graphPopupMenu.add(_saveGraphAction);
		
		_graph.addMouseListener(new GraphMouseListener());
	
		_graph.setMinimumSize(new Dimension(300, 200));
		
		return _graph;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseProgressPanel() {
		
		JPanel panel = new JPanel(new GridLayout(1, 1, 2, 2));
		panel.setBorder(new EmptyBorder(0, 2, 2, 2));
	
		panel.add(_progressComponent);
		
		return panel;
		} // End of method
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseButtonPanel() {
		
		_searchAction = new SearchAction();
		_abortAction = new AbortAction();
		_abortAction.setEnabled(false);
		
		JPanel panel = new JPanel(new GridLayout(1, 3, 5, 5));
		panel.setBorder(new EmptyBorder(5, 5, 5, 5));
		
		JPanel parameterPanel = new JPanel(new GridLayout(1, 1, 0, 0));
		panel.add(parameterPanel);
		
		KeaButton searchButton = new KeaButton(
			_searchAction, 
			KeaButtonColour.Green
			);
		panel.add(searchButton);
		
		KeaButton abortButton = new KeaButton(
			_abortAction, 
			KeaButtonColour.Red
			);
		panel.add(abortButton);
		
		//================================================================
		
		JLabel mismatchSpinnerLabel = new JLabel(" maximum mismatches allowed");
		
		_mismatchSpinner = new JSpinner(
			new SpinnerNumberModel(	
				_maxMismatches,		// Value
				0,					// Min 
				7,					// Max
				1					// step
				)
			);
		_mismatchSpinner.addChangeListener(
			new ChangeListener() {
				public void stateChanged(ChangeEvent ce) {
					mismatchChange();
					}
				}
			);
		
		JTextField tf = ((JSpinner.DefaultEditor)_mismatchSpinner.
			getEditor()).
				getTextField();
		
		tf.setColumns(2);
		tf.setFont(new Font("Dialog", Font.BOLD, 12));
		tf.setForeground(Color.BLUE); 
		
		
		JPanel mismatchSpinnerPanel = new JPanel(new BorderLayout());
		mismatchSpinnerPanel.add(mismatchSpinnerLabel, BorderLayout.CENTER);
		mismatchSpinnerPanel.add(_mismatchSpinner, BorderLayout.WEST);
		
		parameterPanel.add(mismatchSpinnerPanel);
		 
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void mismatchChange() {
		
		// Update member field (ONLY PLACE WHERE FIELD SHOULD BE CHANGED).
		_maxMismatches = ((Integer)_mismatchSpinner.getValue()).intValue();
		
		// Update graph axis.
		if (_maxMismatches == 0) _graph.getXAxis().setTo(1);
		else _graph.getXAxis().setTo(_maxMismatches);
		_graph.repaint();
		
		// Replot if necessary.
		replotGraph();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void replotGraph() {
		
		// Get existing coordinate collections.
		CoordinateCollection[] cc = 
			_graph.getGraphModel().getCoordinateCollections();
		
		// Return if no lines have yet been plotted.
		if (cc == null) {
			return;
			}
		
		// Create new coordinate collections.
		CoordinateCollection[] newCc = new CoordinateCollection[2];
		for (int i = 0; i < cc.length; i++) {
			
			// Get name of plot.
			String s = cc[i].getTitle();
			
			// Determine number of coordinates to plot.
			int numberOfCoordinatesToPlot = 
				java.lang.Math.min(cc[i].size(), (_maxMismatches + 1));
			
			// Declare new coordinates to fill new collection.
			Coordinate[] newCoordinates = 
				new Coordinate[numberOfCoordinatesToPlot];
			
			// Fill new coordinates with old up to max mismatches.
			for (int j = 0; j < numberOfCoordinatesToPlot; j++) {
				newCoordinates[j] = cc[i].getCoordinate(j);
				}
			
			// Initialise new collection.
			newCc[i] = 
				CoordinateCollectionFactory.
					createCoordinateCollection(newCoordinates, s);
			}
		
		_graph.getGraphModel().setCoordinateCollections(newCc);
		setPlotLineFormatters(_graph, newCc);
		_graph.plot();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JMenuBar menuBarComponent() {
		
		JMenuBar menuBar = new JMenuBar();
		menuBar.setBorder(new EmptyBorder(0,0,0,0));
		
		// File
		JMenu fileMenu = new JMenu("File");
		menuBar.add(fileMenu);
		fileMenu.getAccessibleContext().setAccessibleName("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		fileMenu.add(_openDatabaseAction);
		fileMenu.add(new JSeparator());
		fileMenu.add(_oligoPanel.getOpenAction());
		fileMenu.add(_oligoPanel.getSaveAction());
		fileMenu.add(_oligoPanel.getSaveAsAction()); 
		
		// Edit menu.
		menuBar.add(_oligoPanel.getEditMenu());
		
		// Options
		JMenu optionsMenu = new JMenu("Options");
		menuBar.add(optionsMenu);
		optionsMenu.getAccessibleContext().setAccessibleName("Options");
		optionsMenu.setMnemonic(KeyEvent.VK_O);
		optionsMenu.add(_showDataAction);
		optionsMenu.add(_printGraphAction);
		optionsMenu.add(_saveGraphAction);
		optionsMenu.add(new JSeparator());
		optionsMenu.add(this._enableSelectionAction);
		
		// Help.
		_helpMenu = new JMenu("Help");
		menuBar.add(_helpMenu);
		_helpMenu.getAccessibleContext().setAccessibleName("Help");
		_helpMenu.setMnemonic(KeyEvent.VK_H);
		
		return menuBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Use this method to add menu items (Actions) to the help menu.
	 @param action Action to add.
	 */
	public void addHelpMenuAction(final Action action) {
		_helpMenu.add(action);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void abort() {
		_abortParser = true;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void search() {
		
		try {
			_oligo = getOligoFromDisplay();
			} 
		catch(NullSequenceException nse) {
			JOptionPane.showMessageDialog(
				this, 
				"No oligonucleotide has been entered!",
				"Error",
				JOptionPane.ERROR_MESSAGE
				);
			return;
			}
		catch(NonIUBCharacterException nice) {
			JOptionPane.showMessageDialog(
				this, 
				"Oligonucleotide contains one or more illegal base characters.", 
				"Error", 
				JOptionPane.ERROR_MESSAGE
				);
			return;
			}
		catch(IllegalSequenceException ise) {
			JOptionPane.showMessageDialog(
				this, 
				"Oligonucleotide contains both T and U bases!", 
				"Error", 
				JOptionPane.ERROR_MESSAGE
				);
			return;
			}
		
		// Reset:
		//=================================
		_abortParser = false;
		_progressComponent.reset();
		_databaseTree.getTree().setCellRenderer(_primroseTreeCellRenderer);
		_table.setDefaultRenderer(Record.class, _primroseTableCellRenderer);
		
		_oligo.setSearchComplete(false);
		ReportTableModel model = (ReportTableModel)_table.getModel();
		model.updateModel(_oligo);
		model.setColumnWidths(_table);
		
		_graph.clear();
		_showDataAction.setEnabled(false);
		_printGraphAction.setEnabled(false);
		_saveGraphAction.setEnabled(false);
		
		_oligoPanel.getClearAction().setEnabled(false);
		_oligoPanel.getCopyAction().setEnabled(false);
		_oligoPanel.getPasteAction().setEnabled(false);
		_oligoPanel.getDeleteAction().setEnabled(false);
		_oligoPanel.getCutAction().setEnabled(false);
		_oligoPanel.getUndoAction().setEnabled(false);
		_oligoPanel.getRedoAction().setEnabled(false);
		_oligoPanel.getSelectAllAction().setEnabled(false);
		
		// Prevent further alterations to records.
		_databaseTree.enableTreeRecordSelection(false);
		
		//=================================
		
		
		
		final SearchResult maxMismatches = SearchResult.getEnum(
			((Integer)_mismatchSpinner.getValue()).intValue()
			);
		
		final AbortListener abortListener = new AbortListener() {
			public boolean abort() {return _abortParser;} 
			};	
			
		// Initialise SearchDatabase object. 
		final SearchDatabase2 searchDatabase = new SearchDatabase2( 
			this, 
			_database, 
			_oligo,
			_progressComponent,
			maxMismatches, 
			abortListener 
			);
		
		_searchAction.setEnabled(false);
		_abortAction.setEnabled(true);
		_oligoPanel.setEnabled(false);
		_mismatchSpinner.setEnabled(false);
		
		_openDatabaseAction.setEnabled(false);
		
		
		// Run search on separate thread.
		//===================================================================
		Thread t = new Thread(new Runnable(){public void run() {
			
			// DO SEARCH
			searchDatabase.run(); 
			
			// If aborted search:
			if (_abortParser) {_progressComponent.reset();}			
			
			// Successfully completed search.
			else {
				updateDisplays(_oligo);
				_abortParser = true;
				} 
			
			_searchAction.setEnabled(true);
			_abortAction.setEnabled(false);
			_oligoPanel.setEnabled(true);
			_mismatchSpinner.setEnabled(true);
			
			_showDataAction.setEnabled(true);
			_printGraphAction.setEnabled(true);
			_saveGraphAction.setEnabled(true);
			
			// Allow user to select records.
			//_databaseTree.enableTreeRecordSelection(true);
			_enableSelectionAction.setEnabled(true);
			
			// Enable opening of additional sequence files.
			_openDatabaseAction.setEnabled(true);
		
			// Re-enable edit functions associated with oligo panel.
			_oligoPanel.getClearAction().setEnabled(true);
			_oligoPanel.getCopyAction().setEnabled(true);
			_oligoPanel.getPasteAction().setEnabled(true);
			_oligoPanel.getDeleteAction().setEnabled(true);
			_oligoPanel.getCutAction().setEnabled(true);
			_oligoPanel.getUndoAction().setEnabled(true);
			_oligoPanel.getRedoAction().setEnabled(true);
			_oligoPanel.getSelectAllAction().setEnabled(true);
			
			// Ensure dialog queries user on closing.
			_dataGenerated = true;
			
			}});	
		t.setPriority(t.MIN_PRIORITY);
		t.start();
		//====================================================================
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private PrimroseOligo getOligoFromDisplay() 
		throws 
			NonIUBCharacterException,
			NullSequenceException,
			IllegalSequenceException
		{
			
		PrimroseOligo oligo = new PrimroseOligo(
				_oligoPanel.getOligoRecord().getOligo().getSequenceString(),
				_oligoPanel.getOligoRecord().getOrientation()
				);
		return oligo;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Update tree, table, graph and corresponding data table.
	private void updateDisplays(PrimroseOligo oligo) {
		
		/// Tree
		_databaseTree.getTree().setCellRenderer(_reportTreeCellRenderer);
		
		// Table
		_oligo.setSearchComplete(true);
		_table.setDefaultRenderer(Record.class, _reportTableCellRenderer);
		_table.repaint();
		
		// Graph and associated table.
		updateGraph();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private int totalTargets() {
		
		Record[] records = _database.getSelectedRecords();
		int counter = 0;
		for (int i = 0; i < records.length; i++) {
			if (records[i].getRecordType() == RecordType.Target) {
				counter++;
				} 
			}
		
		return counter;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void updateGraph() {
		
		final int SIZE = _maxMismatches + 1;
		
		Record[] records = _oligo.getRecords();
		int[] targetScores = new int[SIZE];
		int[] nonTargetScores = new int[SIZE];
		int total = _database.getSelectedRecords().length;
		int totalTargets = totalTargets();
		int totalNonTargets = total - totalTargets;
		
		// Total up all target scores / non-target scores.
		for (int i = 0; i < records.length; i++) {
	
			// Convert search result to integer.  If no match (i.e. value > 7), 
			// do nothing.
			int score = _oligo.getSearchResult(records[i]).getValue();
			if (score > 7) {continue;}
			
			if (records[i].getRecordType() == RecordType.Target) {
				targetScores[score]++;
				}
			else {
				nonTargetScores[score]++;
				}
			
			}
		
		Coordinate[] targetCoordinates = new Coordinate[SIZE];
		Coordinate[] nonTargetCoordinates = new Coordinate[SIZE];
		
		int target = 0;
		int nonTarget = 0;
		for (int i = 0; i < targetScores.length; i++) {
			
			target += targetScores[i];
			nonTarget += nonTargetScores[i];
			
			if (totalTargets == 0) {
				targetCoordinates[i] = CoordinateFactory.create(i, 0);
				}
			else {
				targetCoordinates[i] = CoordinateFactory.create(
					i, 
					((float)target / (float)totalTargets) * 100
					);
				
				}
			
			if (totalNonTargets == 0) {
				nonTargetCoordinates[i] = CoordinateFactory.create(i, 0);
				}
			else {
				nonTargetCoordinates[i] = CoordinateFactory.create(
					i, 
					((float)nonTarget / (float)totalNonTargets) * 100
					);
				}
			
			}
		
		CoordinateCollection[] collections = new CoordinateCollection[2];
		
		
		collections[0] = 
			CoordinateCollectionFactory.createCoordinateCollection(
				targetCoordinates, "Targets"
				);
		collections[1] = 
			CoordinateCollectionFactory.createCoordinateCollection(
				nonTargetCoordinates, "Non-targets"
				);
		
		_graph.getGraphModel().setCoordinateCollections(collections);
		
		setPlotLineFormatters(_graph, collections);

		_graph.plot();
		
		// Finally update associated table.
		_graphTableDialog.update(collections);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void setPlotLineFormatters(Graph g, CoordinateCollection[] cc) {
		
		PlotLineFormatter plf1 = new DefaultPlotLineFormatter(cc[0]);
		plf1.setGapToSymbol(false); 
		plf1.setLineColor(Color.BLUE); 
		plf1.setLineStyle(LineStyle.Solid); 
		plf1.setLineWidth(1);
		plf1.setPlotStyle(PlotStyle.LineAndScatter); 
		plf1.setSymbol(Symbol.ClosedSquare); 
		plf1.setSymbolColor(Color.BLUE); 
		plf1.setSymbolSize(10); 
		
		PlotLineFormatter plf2 = new DefaultPlotLineFormatter(cc[1]);
		plf2.setGapToSymbol(false); 
		plf2.setLineColor(Color.RED); 
		plf2.setLineStyle(LineStyle.Solid); 
		plf2.setLineWidth(1); 
		plf2.setPlotStyle(PlotStyle.LineAndScatter); 
		plf2.setSymbol(Symbol.ClosedCircle); 
		plf2.setSymbolColor(Color.RED); 
		plf2.setSymbolSize(10); 
		
		g.getCoordinateCollectionRenderer().setPlotLineFormatters(
			new PlotLineFormatter[]{plf1, plf2}
			);	
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void printGraph() {
		
		// Get list of all printers that can handle Printable objects.
		DocFlavor flavor = DocFlavor.SERVICE_FORMATTED.PRINTABLE;
		PrintService[] services = 
			PrintServiceLookup.lookupPrintServices(flavor, null);
		
		// Set some printing attributes.
		PrintRequestAttributeSet printAttributes = 
			new HashPrintRequestAttributeSet();
		printAttributes.add(OrientationRequested.LANDSCAPE);
		
		// Display a dialog that allows the user to select one of the available
		// printers and to edit te default attributes.
		PrintService service = ServiceUI.printDialog(
			null, 
			100, 
			100, 
			services, 
			null, 
			null, 
			printAttributes
			);
		
		// If the user cancelled, don't do anything.
		if (service == null) return;
		
		// Now call a method to finish the printing.
		printToService(service, printAttributes);
		 
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void printToService(
		PrintService service, 
		PrintRequestAttributeSet printAttributes
		) {
		
		DocFlavor flavor = DocFlavor.SERVICE_FORMATTED.PRINTABLE;
		Doc doc = new SimpleDoc(_graph, flavor, null);
			
		DocPrintJob job = service.createPrintJob();
		
		/*
		// Set up a dialog box to monitor printing status.
		final JOptionPane pane = new JOptionPane(
			"Printing...",
			JOptionPane.PLAIN_MESSAGE
			);
			
		JDialog dialog = pane.createDialog(this, "Print Status");
		
		// This listener object updates the dialog as the status changes.
		job.addPrintJobListener(new PrintJobAdapter() {
			public void printJobCompleted(PrintJobEvent pje) {
				//pane.setMessage("Printing complete!");
				}
			public void printDataTransferCompleted(PrintJobEvent pje) {
				//pane.setMessage("Document transferred to printer.");
				}
			public void printJobRequiresAttention(PrintJobEvent pje) {
				pane.setMessage("Check printer: out of paper?");	
				}
			public void printJobFailed(PrintJobEvent pje) {
				pane.setMessage("Print job failed.");
				} 
			});
		
		// Show the dialog. Non-modal.
		dialog.setModal(false);
		dialog.show();
		*/
		
		// Now print the Doc to the DocPrintJob.
		try {
			job.print(doc, printAttributes);
			} 
		
		catch (PrintException pe) {
			JOptionPane.showMessageDialog(
				this, 
				"Unable to print.", 
				"Print Error", 
				JOptionPane.ERROR
				);
			}
			
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void saveGraph() throws IOException {
		
		JpegFileFilter _jpegFileFilter = new JpegFileFilter();
		PngFileFilter _pngFileFilter = new PngFileFilter();
		PostScriptFileFilter _postScriptFileFilter = new PostScriptFileFilter();
		
		ResultsDialogFileChooser fc = new ResultsDialogFileChooser();
		
		// Remove AcceptAll file filter.
		fc.removeChoosableFileFilter(fc.getAcceptAllFileFilter());	
		
		// Add new file filter.
		fc.addChoosableFileFilter(_jpegFileFilter);
		fc.addChoosableFileFilter(_pngFileFilter);
		fc.addChoosableFileFilter(_postScriptFileFilter);
		
		// Display current file, if any.
		if (_graphFile != null) {
			fc.setSelectedFile(
				new File(FileIO.getNameWithoutSuffix(_graphFile))
				);
			}
		
		fc.setDialogTitle("Save graph image");
		if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
			
			// Set file.
			_graphFile = fc.getSelectedFile();
			
			// save in appropriate format.
			if (fc.getFileFilter() == _jpegFileFilter) {
				saveGraphAsJPEG(_graphFile);
				}
			else if (fc.getFileFilter() == _pngFileFilter) {
				saveGraphAsPNG(_graphFile);
				}
			
			else if (fc.getFileFilter() == _postScriptFileFilter) {
				saveGraphAsPostScript(_graphFile);
				}
			
			else {
				throw new IllegalArgumentException(
					"Could not identify correct FileFilter: " + 
					fc.getFileFilter().getDescription()
					);
				}
			
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void saveGraphAsJPEG(File file) throws IOException {
		
		RenderedImage image = _graph.createImage();
		ImageIO.write(
			image, 
			"jpeg", 
			file//FileIO.ensureCorrectFileExtension(file, new JpegFileFilter())
			);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void saveGraphAsPNG(File file) throws IOException {
		
		RenderedImage image = _graph.createImage();
		ImageIO.write(
			image, 
			"png", 
			file//FileIO.ensureCorrectFileExtension(file, new PngFileFilter())
			);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void saveGraphAsPostScript(File file) throws IOException {
		
		DocFlavor flavor = DocFlavor.SERVICE_FORMATTED.PRINTABLE;
		String format = "application/postscript";
		
		StreamPrintServiceFactory factory = 
			StreamPrintServiceFactory.
				lookupStreamPrintServiceFactories(flavor, format)[0];
		
		FileOutputStream out = new FileOutputStream(
			file//FileIO.ensureCorrectFileExtension(file, new PostScriptFileFilter())
			);
		
		// Obtain a PrintService that prints to that file.
		StreamPrintService service = factory.getPrintService(out);
		
		// Do printing.
		printToService(service, null);
		
		out.close();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// User confirms via dialog box that it is ok (or not) to overwrite 
	// existing file.
	private boolean okToOverwrite(String filename) {
		
		if (JOptionPane.showConfirmDialog(
			this,
			"'" + filename + "' already exists.  Do you want to replace it?",
			"Warning",
			JOptionPane.YES_NO_OPTION,
			JOptionPane.INFORMATION_MESSAGE
			) == JOptionPane.YES_OPTION) {return true;} else {return false;}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Fills table with data supplied by Record object.  Fired by expansion or
	// selection of treenode.  Also fired when selection status is toggled, and
	// when table record is double clicked.
	private void updateTable(Record record) {
		
		// Do nothing if a leaf.
		if (record.isLeaf()) {return;}
		
		// Get model associated with table.
		ReportTableModel m = (ReportTableModel)_table.getModel();
			
		// Display new data.
		m.updateModel(record, _oligo);

		// Reset column sizes (otherwise all same size).
		m.setColumnWidths(_table);

		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Ensures expanded status of tree is up-to-date with table.  When a table
	// record is double clicked, and the children of that record displayed in 
	// the table, so should the tree be expanded to reflect this change.
	private void expandTree(Record record) {
		
		// Procedure: Get all parents of record.  Create an array from these
		// Record objects.  Use this array to create a TreePath object.  Use
		// this TreePath object to expand tree.
		
		// Stack to store record and parents of record.
		Stack stack = new Stack();
		stack.push(record);
		
		// Store parents in stack (order: most recent parent to root).
		Record parent = null;
		while ((parent = (Record)record.getParent()) != null) {
			stack.push(parent);
			record = parent;
			}
		
		// Now create array and fill it with stack contents in reverse order. 
		int size = stack.size();
		Record[] parents = new Record[size];
		
		int counter = 0;
		while(!stack.empty()) {
			parents[counter] = (Record)stack.pop();
			counter++;
			}
		
		// Create TreePath object and use to expand tree.
		TreePath tp = new TreePath(parents);
		_databaseTree.getTree().expandPath(tp);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private boolean okToClose() {
		
		// No search running and no data displayed.
		if (_abortParser && !_dataGenerated) {
			return true;
			}
		
		// Search running.
		if (!_abortParser) {
			
			int userResponse = 
				JOptionPane.showConfirmDialog(
					this, 
					"<html>This will cancel your search.<p>" + 
					"Are you sure you want to do this?</html>",
					"Closing...",
					JOptionPane.YES_NO_OPTION
					);
			
			if (userResponse == JOptionPane.YES_OPTION) {
				return true;
				} 
			else {
				return false;
				}
			
			}
		
		// Data displayed.
		else {
			int userResponse = 
				JOptionPane.showConfirmDialog(
					this, 
					"<html>Results will be lost if you close this window.<p>" + 
					"Do you want to continue?</html>",
					"Closing...",
					JOptionPane.YES_NO_OPTION
					);
			
			if (userResponse == JOptionPane.YES_OPTION) {
				return true;
				} 
			else {
				return false;
				}
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void notepad() {
		
		final SimpleNotepadDialog snd =
			new SimpleNotepadDialog(this, true, _oligoPanel.getNotes()); 
			
		snd.setBackground(Color.WHITE);
		snd.setForeground(Colour.NavyBlue);
		snd.setFont(new Font("Monospaced", Font.PLAIN, 14));
		snd.setUndecorated(false);
		
		if (DialogResult.Ok == snd.showDialog()) {
			_oligoPanel.setNotes(snd.getText());
			}
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/** 
	 Overrides base class method if the WINDOW_CLOSING window event is fired.
	 @param we Event passed from EventListener.
	 */
	protected void processWindowEvent(WindowEvent we) {
		
		// Action if window closing exit is being fired.
		if (we.getID() == WindowEvent.WINDOW_CLOSING) {
			
			if (okToClose()) {
				if (!_abortParser) {abort();} // abort parser if still running.
				setVisible(false);
				dispose();
				}
		
			} 
		// Otherwise use base class method.
		else {
			super.processWindowEvent(we);
			}
		
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	private class TreeMouseListener implements MouseListener {
		
		public void mouseClicked(MouseEvent me) {
			if (me.isPopupTrigger()) {selectRecordToggle(me);}
			}
		
		public void mouseEntered(MouseEvent me) {
			}
		
		public void mouseExited(MouseEvent me) {
			}
		
		public void mousePressed(MouseEvent me) {
			if (me.isPopupTrigger()) {selectRecordToggle(me);}
			}
		
		public void mouseReleased(MouseEvent me) {
			if (me.isPopupTrigger()) {selectRecordToggle(me);}
			}
		
		//=====================================================
		
		private void selectRecordToggle(MouseEvent me) {
			
			TreePath tp = 
				_databaseTree.
					getTree().
						getClosestPathForLocation(me.getX(), me.getY());
			
			if (tp == null) {return;}
			
			//Record r = (Record)tp.getLastPathComponent();
			//updateTable(r);
			
			_table.repaint();
			
			} // End of inner class method.
		
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Open record within table and expand tree.
	private class TableMouseListener extends MouseAdapter {
		
		public void mouseClicked(MouseEvent me) {
			if (me.getClickCount() == 2) {openRecord(me);}
			}
		
		private void openRecord(MouseEvent me) {
			
			int rowIndex = _table.rowAtPoint(me.getPoint());
			int columnIndex = _table.columnAtPoint(me.getPoint());
			
			Object o = _table.getModel().getValueAt(rowIndex, columnIndex);
			
			if (o instanceof Record) {
				Record r = (Record)o;
				updateTable(r);
				expandTree(r);
				}
			
			} 
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Update table when tree changes.
	private class SelectionListener implements TreeSelectionListener {
		
		public void valueChanged(TreeSelectionEvent e) {
			Record r = (Record)e.getPath().getLastPathComponent();
			updateTable(r);
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	// Update table when tree changes.
	private class ExpansionListener implements TreeExpansionListener {
		
		public void treeCollapsed(TreeExpansionEvent e) {
			
			}
		
		public void treeExpanded(TreeExpansionEvent e) {
			Record r = (Record)e.getPath().getLastPathComponent();
			
			updateTable(r);
			
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class GraphMouseListener extends MouseAdapter {
		
		public void mouseClicked(MouseEvent me) {
			if (me.isPopupTrigger()) {popupMenu(me);}
			}
		
		public void mousePressed(MouseEvent me) {
			if (me.isPopupTrigger()) {popupMenu(me);}
			}
		public void mouseReleased(MouseEvent me) {
			if (me.isPopupTrigger()) {popupMenu(me);}
			}
		
		private void popupMenu(MouseEvent me) {
		
			_graphPopupMenu.show(
					_graph, 
					me.getX(), 
					me.getY()
					);
			
			} 
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class SearchAction extends AbstractAction {
		
		public SearchAction() {
			putValue(Action.NAME, "Search Database Records");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Search database records with oligonucleotide."
				);
			putValue(Action.SMALL_ICON, KeaIcon.AnalyseDatabase.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent ae) {
			search();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class EnableSelectionAction extends AbstractAction {
		
		public EnableSelectionAction() {
			putValue(Action.NAME, "Enable Record Selection/Deselection");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Allows sequence records to be selected or de-selected."
				);
			putValue(
				Action.SMALL_ICON, 
				KeaIcon.EnableRecordSelection.getImageIcon()
				);
			}
		
		public void actionPerformed(ActionEvent ae) {
			
			// Allow user to select records.
			_databaseTree.enableTreeRecordSelection(false); // re bug.
			_databaseTree.enableTreeRecordSelection(true);
			
			// clear results (to avoid error messages).
			_oligo.setSearchComplete(false);
			_databaseTree.getTree().setCellRenderer(_primroseTreeCellRenderer);
			_table.setDefaultRenderer(Record.class, _primroseTableCellRenderer);
			_table.repaint();
			
			// Disable button - no longer required.
			this.setEnabled(false);
			
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class AbortAction extends AbstractAction {
		
		public AbortAction() {
			putValue(Action.NAME, "Abort Search");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Cancels current database search."
				);
			putValue(Action.SMALL_ICON, KeaIcon.Cancel.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			abort();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class ShowDataAction extends AbstractAction {
		
		public ShowDataAction() {
			putValue(Action.NAME, "Show Data");
			putValue(Action.SHORT_DESCRIPTION, "Displays graph data.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_D));
			putValue(
				Action.ACCELERATOR_KEY,
				KeyStroke.getKeyStroke(KeyEvent.VK_D, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(Action.SMALL_ICON, KeaIcon.Table.getImageIcon());
			} 
		
		public void actionPerformed(ActionEvent e) {
			
			_dataTableVisible = !_dataTableVisible;
			_graphTableDialog.setVisible(_dataTableVisible);
			
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
		private class PrintGraphAction extends AbstractAction {
		
		public PrintGraphAction() {
			putValue(Action.NAME, "Print Graph...");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Prints graph.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_P));
			putValue(
				Action.ACCELERATOR_KEY,
				KeyStroke.getKeyStroke(KeyEvent.VK_P, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(
				Action.SMALL_ICON, 
				kea.programs.ProgramInfo.getPrintIcon()
				);
			}
		
		public void actionPerformed(ActionEvent e) {
			printGraph();
			}		
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class SaveGraphAction extends AbstractAction {
		
		public SaveGraphAction() {
			putValue(Action.NAME, "Save Graph...");
			putValue(Action.SHORT_DESCRIPTION, "Saves graph image.");
			putValue(Action.MNEMONIC_KEY, new Integer(KeyEvent.VK_G));
			putValue(
				Action.ACCELERATOR_KEY,
				KeyStroke.getKeyStroke(KeyEvent.VK_G, InputEvent.CTRL_DOWN_MASK)
				);
			putValue(Action.SMALL_ICON, KeaIcon.SaveGraph.getImageIcon());
			} 
		
		public void actionPerformed(ActionEvent ae) {
			try {
				saveGraph();
				}
			catch(IOException ioe) {
				JOptionPane.showMessageDialog(
					Report.this, 
					"Unable to save graph to file.", 
					"IOException Error", 
					JOptionPane.ERROR_MESSAGE);
				}
			catch (Exception e) {
				e.printStackTrace();
				}
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class ResultsDialogFileChooser extends JFileChooser {
		
		
		// Override method.
		public void approveSelection() {
			
			// Make sure filename supplied by the user has the correct
			// ending.
			File file = FileIO.ensureCorrectFileExtension(
				this.getSelectedFile(), 
				(KeaFileFilter)this.getFileFilter()
				);
			
			// Check that no file is being unintentionally overwritten.
			File directory = super.getCurrentDirectory();
			if (this.okToSave(file, directory)) {
				super.approveSelection();
				this.setSelectedFile(file);
				}
			else {return;}
			
			}
		
		// Checks that no file is being overwritten.
		private boolean okToSave(File file, File directory) {
			
			// Get list of files in current directory.
			File[] files = directory.listFiles((KeaFileFilter)this.getFileFilter());

			// Check user supplied name with these.
			for (int i = 0; i < files.length; i++) {

				String name1 = files[i].getName();
				String name2 = file.getName();

				if (name1.equals(name2)) {
					return okToOverwrite(name1);
					}
				}

			// No existing file with same name therefore ok to save.
			return true;

			} // End of method.
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class NotepadAction extends AbstractAction {
		
		public NotepadAction() {
			putValue(Action.NAME, "Notepad");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Notes typed here will be saved if the oligonucleotide is saved."
				);
			putValue(
				Action.ACCELERATOR_KEY, 
				KeyStroke.getKeyStroke(KeyEvent.VK_F5, 0)
				);
			putValue(Action.SMALL_ICON, KeaIcon.Notepad.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			notepad();
			}
		
		} // End of inner class.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////