// ReportGraphTableModel.java created on 29 May 2004 at 10:45

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import javax.swing.table.*;
import java.awt.*;
import javax.swing.*;
import java.util.*;

import kea.programs.plot.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Table model for report graph-table.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public final class ReportGraphTableModel extends AbstractTableModel {
	
	//##########################################################################
	
	// MEMBER FIELDS

	private final String[] _columnNames;
	private final Class[] _columnClasses;
	private CoordinateCollection[] _coordinateCollections;
	
	private static final String COLUMN_1_NAME = "Mismatches";
	private static final String COLUMN_2_NAME = "% Within target";
	private static final String COLUMN_3_NAME = "% Outside target";
	
	private static final int MAX_ROWS = 8; // i.e. 0 - 7 mismatches.
	
	/**
	 Null data for displaying in table.
	 */
	public static final CoordinateCollection[] NULL_DATA;
	
	// Initialiser block.
	static {
		
		// Create 'null' data for initial display.
		Coordinate[] c1 = new Coordinate[]{};	
			
		// Fill array.	
		NULL_DATA = new CoordinateCollection[] {
			CoordinateCollectionFactory.createCoordinateCollection(
				c1, "EMPTY1"
				),
			CoordinateCollectionFactory.createCoordinateCollection(
				c1, "EMPTY2"
				)
			};	
		
		}
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of class.
	 */
	public ReportGraphTableModel() {

		_columnNames = new String[]{
			COLUMN_1_NAME, 
			COLUMN_2_NAME,
			COLUMN_3_NAME
			};
		_columnClasses = new Class[] {
			String.class,
			String.class,
			String.class
			};	
		
		updateModel(NULL_DATA);	
		
		} // End of constructor.
	
	//##########################################################################
	
	// METHODS

	/**
	 Update model with new data for display.
	 @param collections Array of CoordinateCollection objects containing data
	 to display.
	 */
	public void updateModel(CoordinateCollection[] collections) {
		_coordinateCollections = collections;
		// Force new data to be displayed.
		fireTableChanged(null);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Implementation of AbstractTableModel method.
	 @return Number of columns.
	 */
	public int getColumnCount() {
		return _columnClasses.length;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Implementaion of AbstractTableModel method.
	 @return Number of rows.
	 */
	public int getRowCount() {
		return MAX_ROWS;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Implementation of AstractTableModel method.
	 @param rowIndex Row index number for cell.
	 @param columnIndex Column index number for cell.
	 @return Object.
	 */
	public Object getValueAt(int rowIndex, int columnIndex) {
		
		switch (columnIndex) {
			case 0: return String.valueOf(rowIndex);
			case 1: return getData(0, rowIndex);
			case 2: return getData(1, rowIndex);
			default: return null;
			} // End of switch.		
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////

	private String getData(final int collectionNumber, final int rowIndex) {
		
		CoordinateCollection cc = _coordinateCollections[collectionNumber];
	
		// If no data available return blank string.
		if (rowIndex > cc.size() - 1) {return "-";} 
		
		// Otherwise, return data.
		else {
			return formatNumber(cc.getCoordinate(rowIndex).getY());
			}
	
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Tidy up float numbers for display.
	private String formatNumber(float number) {
		
		java.text.NumberFormat nf = java.text.NumberFormat.getInstance();
		
		nf.setMinimumFractionDigits(1);
		nf.setMaximumFractionDigits(1);
		
		return nf.format(number);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Overridden AbstractTableModel method.
	 @param columnIndex Column index number.
	 @return Column name as a String.
	 */
	public String getColumnName(int columnIndex) {
		return _columnNames[columnIndex];
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Overridden AbstractTableModel method.
	 @param columnIndex Column index number.
	 @return Class object for column.
	 */
	public Class getColumnClass(int columnIndex) {
		return _columnClasses[columnIndex];
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Sets widths of columns.
	 @param table JTable object within which columns will be displayed.
	 */
	public void setColumnWidths(JTable table) {
		
		TableColumnModel model = table.getColumnModel();
		
		// NOTE:
		// Done the following way to avoid complications if/when new columns
		// are added and the order changed.
		
		TableColumn column = null;
		String name = null;
		Enumeration columns = model.getColumns();
		while(columns.hasMoreElements()) {
			
			column = ((TableColumn)columns.nextElement());
			name = ((String)column.getHeaderValue());
			
			if (name.equals(COLUMN_1_NAME)) {
				column.setPreferredWidth(30);
				}
			
			if (name.equals(COLUMN_2_NAME)) {
				column.setPreferredWidth(85);
				}
			
			if (name.equals(COLUMN_3_NAME)) {
				column.setPreferredWidth(85);
				}
			
			} // End of while loop.
		
		} // End of method.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////