// ReportTableModel.java created on 26 May 2004 at 17:54

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import java.util.*;
import javax.swing.table.*;
import javax.swing.*;
import java.io.*;

import kea.database.*;
import kea.sequences.nucleicAcid.search.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Table model for Report table.
 @author  Kevin Ashelford.
 @version 2.17.
 */
class ReportTableModel extends AbstractTableModel {
	
	//##########################################################################
	
	// MEMBER FIELDS

	private String[] _columnNames;
	private Class[] _columnClasses;
	private Vector _children;
	private Record _record;
	private PrimroseOligo _oligo;
	
	private static final String COLUMN_1_NAME = "Name";
	private static final String COLUMN_2_NAME = "Code";
	private static final String COLUMN_3_NAME = "Length";
	private static final String COLUMN_4_NAME = "Result";
	private static final String COLUMN_5_NAME = "Location";
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of class.
	 @param record Record object that will supply data.
	 @param numberOfOligos Number of oligos under consideration.
	 */
	public ReportTableModel(Record record, PrimroseOligo oligo) {
		
		_columnNames = new String[]{
			COLUMN_1_NAME, 
			COLUMN_2_NAME,
			COLUMN_3_NAME,
			COLUMN_4_NAME,
			COLUMN_5_NAME
			};
			
		_columnClasses = new Class[] {
			Record.class,
			String.class,
			String.class,
			String.class,
			String.class
			};
	
		updateModel(record, oligo);	
		
		} // End of constructor.
	
	//##########################################################################

	// METHODS
	
	/**
	 Update model with new Record and Oligo object.
	 @param record Record object containing data to display in table.
	 @param oligo Oligo object containing information relevant to Record.
	 */
	public void updateModel(Record record, PrimroseOligo oligo) {
		_record = record;
		_oligo = oligo;
		_children = _record.getChildren();
		fireTableChanged(null);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Update model with new Oligo object only.
	 @param oligo Oligo object.
	 */
	public void updateModel(PrimroseOligo oligo) {
		updateModel(_record, oligo);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Returns current Record object (supplying data) associated with model.
	 @return Record object.
	 */
	public Record getCurrentRecord() {
		return _record;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Gets source name from File object associated with Record object.
	private String getSourceName(Record record) {
		return kea.util.FileIO.getNameWithoutSuffix(record.getSource());
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Implementation of AbstractTableModel method.
	 @return Number of columns.
	 */
	public int getColumnCount() {
		return _columnClasses.length;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Implementaion of AbstractTableModel method.
	 @return Number of rows.
	 */
	public int getRowCount() {
		return _children.size();
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Implementation of AstractTableModel method.
	 @param rowIndex Row index number for cell.
	 @param columnIndex Column index number for cell.
	 @return Object.
	 */
	public Object getValueAt(int rowIndex, int columnIndex) {
		
		Record child = (Record)_children.get(rowIndex);
		
		// Cater for non-selected records
		if (child.getSelectionStatus() == RecordStatus.NotSelected) {
			return getValueFromNodeChild(child, columnIndex);
			}
		
		// Cater for not relevant records.
		if (child.getMatchStatus() == RecordStatus.NotRelevant) {
			return getValueFromNodeChild(child, columnIndex);
			}
		
		// If child is a leaf:
		if (child.isLeaf()) {return getValueFromLeafChild(child, columnIndex);}
		
		// If child is a node:
		else {return getValueFromNodeChild(child, columnIndex);}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////

	// Identifies correct object for getValueAt() method.
	private Object getValueFromLeafChild(Record child, int columnIndex) {
		
		switch (columnIndex) {
			case 0: return child;
			case 1: return child.getCode();
			case 2: return sequenceLength(child); 
			case 3: return searchResult(child); 
			case 4: return positionInfo(child);
			default: return null;
			} // End of switch.			

		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private String sequenceLength(Record record) {
		
		if (_oligo.searchComplete()) {
			return String.valueOf(_oligo.getSequenceLength(record));
			} 
		else {
			return " ";
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private String searchResult(Record record) {
		
		if (_oligo.searchComplete()) {
			return _oligo.getSearchResult(record).toString();
			} 
		else {
			return " ";
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private String positionInfo(Record record) {
		
		if (_oligo.searchComplete()) {
			
			int start = _oligo.getStartPosition(record);
			int end = _oligo.getEndPosition(record);

			if (_oligo.getSearchResult(record) == SearchResult.NoMatch) {
				return " ";
				}
			else {
				return "" + start + ", " + end;
				}
			
			} 
		else {
			return " ";
			}
		
		
	
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Identifies correct object for getValueAt() method.
	private Object getValueFromNodeChild(Record child, int columnIndex) {
		
		switch (columnIndex) {
			case 0: return child;
			case 1: return child.getCode(); 
			case 2: return " ";
			case 3: return " ";
			case 4: return " ";
			default: return null;
			} // End of switch.			
			
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Overridden AbstractTableModel method.
	 @param columnIndex Column index number.
	 @return Column name as a String.
	 */
	public String getColumnName(int columnIndex) {
		return _columnNames[columnIndex];
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Overridden AbstractTableModel method.
	 @param columnIndex Column index number.
	 @return Class object for column.
	 */
	public Class getColumnClass(int columnIndex) {
		return _columnClasses[columnIndex];
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Sets widths of columns.
	 @param table JTable object within which columns will be displayed.
	 */
	public void setColumnWidths(JTable table) {
		
		TableColumnModel model = table.getColumnModel();
		
		// NOTE:
		// Done the following way to avoid complications if/when new columns
		// are added and the order changed.
		
		TableColumn column = null;
		String name = null;
		Enumeration columns = model.getColumns();
		while(columns.hasMoreElements()) {
			
			column = ((TableColumn)columns.nextElement());
			name = ((String)column.getHeaderValue());
			
			if (name.equals(COLUMN_1_NAME)) {
				column.setPreferredWidth(200);
				}
			
			if (name.equals(COLUMN_2_NAME)) {
				column.setPreferredWidth(40);
				}
			
			if (name.equals(COLUMN_3_NAME)) {
				column.setPreferredWidth(40);
				}
			
			if (name.equals(COLUMN_4_NAME)) {
				column.setPreferredWidth(55);
				}
			
			if (name.equals(COLUMN_5_NAME)) {
				column.setPreferredWidth(45);
				}
			
			} // End of while loop.
		
		} // End of method.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
