// SearchDatabase.java created on 17 May 2004 at 14:55

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import java.io.*;
import javax.swing.*;
import java.awt.*;
import java.util.*;
import java.text.ParseException;
import java.util.regex.*;

import kea.database.*;
import kea.programs.progress.ProgressComponent;
import kea.sequences.nucleicAcid.*;
import kea.io.*;
import kea.exceptions.*;
import kea.parsers.fasta.*;
import kea.parsers.genBank.*;
import kea.event.*;
import kea.programs.database.DatabaseFileChooser;
import kea.sequences.nucleicAcid.search.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Searches database with supplied oligonucleotides.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public class SearchDatabase {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private final Database _database;
	private final PrimroseOligo[] _oligos;
	private final ProgressComponent _progressComponent;
	private final JLabel _progressMessage;
	private final int _allowedNonTargetHits;
	private final Window _owner;
	
	private AbortListener _abortListener;
	private int _failures;
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of <code>SearchDatabase</code>.
	 @param owner Owning window component.
	 @param database Record database to search.
	 @param oligos Oligo records to use in search.
	 @param progressComponent ProgressComponent object to update.
	 @param allowedNonTargetHits Maximum number of non-target hits allowed.
	 @param abortListener Register abort listener object.
	 */
	public SearchDatabase(
		final Window owner,
		final Database database,  
		final PrimroseOligo[] oligos,
		final ProgressComponent progressComponent,
		final int allowedNonTargetHits,
		final AbortListener abortListener
		) {
		this(
			owner, 
			database, 
			oligos, 
			progressComponent, 
			new JLabel(), 
			allowedNonTargetHits, 
			abortListener
			);
		} // End of constructor.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Creates a new instance of <code>SearchDatabase</code>.
	 @param owner Owning window component.
	 @param database Record database to search.
	 @param oligos Oligo records to use in search.
	 @param progressComponent ProgressComponent object to update.
	 @param progressMessage JLabel object to update.
	 @param allowedNonTargetHits Maximum number of non-target hits allowed.
	 @param abortListener Register abort listener object.
	 */
	public SearchDatabase(
		final Window owner,
		final Database database,  
		final PrimroseOligo[] oligos,
		final ProgressComponent progressComponent,
		final JLabel progressMessage,
		final int allowedNonTargetHits,
		final AbortListener abortListener
		) {
		_owner = owner;
		_database = database;
		_oligos = oligos;
		_progressComponent = progressComponent;
		_progressMessage = progressMessage;
		_allowedNonTargetHits = allowedNonTargetHits;
		
		_abortListener = abortListener;

		} // End of constructor.
	
	//##########################################################################
	
	// PROPERTY ACCESSOR METHODS
	
	//##########################################################################
	
	// METHODS
	
	/**
	 Runs search, generating a collection of OligoSearchResult objects.
	 */
	public void run() {	

		_failures = 0;
		
		// Work through each file contributing to database object.
		Iterator iterator = _database.getFileIterator();
		while(iterator.hasNext()) {
			
			
			
			DatabaseFile file = (DatabaseFile)iterator.next();
			
			// Get list of all selected records in file.
			Record[] records = _database.getSelectedRecordsFromFile(file);
			
			// Parse records.
			parseAccordingToFileType(file, records);
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/*
	 
	 Method called by parser handlers - each time a sequence is extracted from
	 a file.
	 
	 NOTE: THIS METHOD NEEDS TO BE AS FAST AS POSSIBLE.
	 
	 */
	
	private void search(final Record record, final String rawSequence) {
		
		String sequence = 
			SequenceConvertor.
				convertToStringWithoutGaps(rawSequence).
					toUpperCase();
		
		// Convert sequence to DNA.
		sequence = SequenceConvertor.convertTo(NucleicAcidType.DNA, sequence);
		
		// For each oligo, run search and store results.
		for (int i = 0; i < _oligos.length; i++) {
			
			// Do not action oligo if exceeded number of non-target hits.
			if (_oligos[i].hasFailed()) {continue;}
			
			// Abort if all failures.
			if (_failures == _oligos.length) {
				/* TODO*/
				}
			
			// Increment target / non-target score (note, in order to avoid
			// OutOfMemoryError errors, record and result is not stored for
			// future use).
			_oligos[i].incrementScore(record, simpleSearch(_oligos[i], sequence));
			
			if (_oligos[i].getNonTargetHits() > _allowedNonTargetHits) {
				_oligos[i].setFailed(true);
				_failures++;  
				_progressMessage.setText(
					_failures + 
					" of out " + 
					_oligos.length + 
					" oligonucleotides have exceeded the non-target limit."
					);
				
				}
			
			} // End of for loop - no more oligos.
		
		} // End of method.
	
	
	////////////////////////////////////////////////////////////////////////////
	
	private SearchResult simpleSearch(final Oligo oligo, final String sequence) {
		
		if (oligo.getMatcher(sequence).find())
			return SearchResult.Match; else return SearchResult.NoMatch;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Identifies file selected and if ok, selects appropriate parsing code.
	private void parseAccordingToFileType(
		final DatabaseFile file, 
		final Record[] records
		) {

		_progressComponent.resetTimer();	
			
		try {
		
			if (file.getFileType() == FileType.RDP_PHYLO) {
				parseFastaFile(file, records);
				}
		
			else if (file.getFileType() == FileType.GENBANK) {
				parseGenBank(file, records);
				}
		
			else if (file.getFileType() == FileType.FASTA) {
				parseFastaFile(file, records);
				}
		
			}
		
		catch(RecordConflictException rce) {
			JOptionPane.showMessageDialog(
				_owner, 
				rce.getMessage(), 
				"Record Conflict Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(FileNotFoundException fnfe) {
			JOptionPane.showMessageDialog(
				_owner, 
				fnfe.toString(),
				"File Not Found Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(IOException ioe) {
			JOptionPane.showMessageDialog(
				_owner, 
				ioe.toString(),
				"IOException Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(DatabaseParserException dpe) {
			JOptionPane.showMessageDialog(
				_owner, 
				dpe.toString(),
				"Database Parser Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(OutOfMemoryError oome) {
			JOptionPane.showMessageDialog(
				_owner, 
				"<html>" + 
				"The Java virtual machine has run out of memory.<p>" + 
				"Suggestion: try re-running the program from the command<br>" + 
				"line, using the -Xmx option to increase heap memory allocation." + 
				"</html>",
				"Out Of Memory Error", 
				JOptionPane.ERROR_MESSAGE
				);
			}
		
		catch(Exception e) {
			JOptionPane.showMessageDialog(
				_owner, 
				e.toString(),
				"Exception Error", 
				JOptionPane.ERROR_MESSAGE
				);
			e.printStackTrace();
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void parseFastaFile(final DatabaseFile file, final Record[] records) 
		throws
			IOException,
			FileNotFoundException,
            DatabaseParserException,
			RecordConflictException,
			NonIUBCharacterException
		{
				
		// If record source is an RDP .phylo file, get the correct GenBank file.
		DatabaseFile seqFile;
		if (FileType.RDP_PHYLO.isFileType(file)) {
			seqFile = getCorrectRDPfastaFile(file);
			} 
		else {
			seqFile = file;
			}		
			
		// Initialise handler.
		MyFastaFileHandler handler = new MyFastaFileHandler(file);
		
		// Create parser object.
		FastaFileParserFactory f = FastaFileParserFactory.newInstance();
		FastaSelectiveParser p = f.newSelectiveParser();
		
		// Register a Progress Listener to update progress bar.
		p.addProgressListener(_progressComponent);
		
		// Register an AbortListener to abort parser if necessary.
		p.addAbortListener(_abortListener);
		
		
		// Construct a list of accession numbers to pass to parser.
		ArrayList parseList = new ArrayList();
		for (int i = 0; i < records.length; i++) {
			parseList.add(records[i].getCode());
			}
		
		// Parse data.	
		p.parse(seqFile, handler, parseList);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void parseGenBank(final DatabaseFile file, final Record[] records) 
		throws
			IOException,
            FileNotFoundException,
            ParseException,
			DatabaseParserException,
			RecordConflictException,
			NonIUBCharacterException
			{
		
		// If record source is an RDP .phylo file, get the correct GenBank file.
		DatabaseFile seqFile;
		if (FileType.RDP_PHYLO.isFileType(file)) {
			seqFile = getCorrectRDPgbFile(file);
			} 
		else {
			seqFile = file;
			}	
		
		// Initialise handler.
		MyGenBankFileHandler handler = new MyGenBankFileHandler(file);
		
		// Create parser object. 
		GenBankFileParserFactory f = GenBankFileParserFactory.newInstance();
		GenBankSelectiveParser p = f.newSelectiveParser(); 
		
		// Register a Progress Listener to update progress bar.
		p.addProgressListener(_progressComponent);
		
		// Register an AbortListener to abort parser if necessary. 
		p.addAbortListener(_abortListener);
		
		// Construct a list of accession numbers to pass to parser.
		ArrayList parseList = new ArrayList();
		for (int i = 0; i < records.length; i++) {
			parseList.add(records[i].getCode());
			}	
			
		// Parse data.
		p.parse(seqFile, handler, parseList);
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Gets RDP .fas file.  If can't find it, asks user for location.
	private DatabaseFile getCorrectRDPfastaFile(final DatabaseFile file) {
		
		// Guess correct path.
		String filepath = file.getAbsolutePath().replaceAll("\\.phylo", ".fas");
		
		// Create DatabaseFile object.
		DatabaseFile dFile = new DatabaseFile(filepath);
		
		// Return DatabaseFile object if file exists.
		if (dFile.isFile()) {return dFile;}
		
		// If does not exist, get user to find it.
		if (
			JOptionPane.showConfirmDialog(
				_owner, 
				"<html>Can't find " + 
				filepath + ".<br>" +
				"Do you want to find <b>" + dFile.getName() + "</b> yourself?", 
				"File error", 
				JOptionPane.YES_NO_OPTION, 
				JOptionPane.ERROR_MESSAGE
				)
				== JOptionPane.OK_OPTION
			) {
			// Initialise dialog.	
			DatabaseFileChooser fc = new DatabaseFileChooser();
			
			// Remove AcceptAll file filter.
			fc.removeChoosableFileFilter(fc.getAcceptAllFileFilter());
			
			// Add database file filter.
			fc.addChoosableFileFilter(new FastaFileFilter());
		
			// Set title.
			fc.setDialogTitle("Open database file");
			
			try {
				
				// Open file selected by user.
				if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
					return fc.getSelectedDatabaseFile();
					} 
				
				} 
			
			catch(FileTypeException fte) {
				
				JOptionPane.showMessageDialog(
					_owner, 
					"Cannot open file - has an unrecognised suffix", 
					"File suffix error", 
					JOptionPane.ERROR_MESSAGE
					);
				}
			}

		return null;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	// Gets RDP .gb file.  If can't find it, asks user for location.
	private DatabaseFile getCorrectRDPgbFile(final DatabaseFile file) {
		
		// Guess correct path.
		String filepath = file.getAbsolutePath().replaceAll("\\.phylo", ".gb");
		
		// Create DatabaseFile object.
		DatabaseFile dFile = new DatabaseFile(filepath);
		
		// Return DatabaseFile object if file exists.
		if (dFile.isFile()) {return dFile;}
		
		// If does not exist, get user to find it.
		if (
			JOptionPane.showConfirmDialog(
				_owner, 
				"<html>Can't find " + 
				filepath + ".<br>" +
				"Do you want to find <b>" + dFile.getName() + "</b> yourself?", 
				"File error", 
				JOptionPane.YES_NO_OPTION, 
				JOptionPane.ERROR_MESSAGE
				)
				== JOptionPane.OK_OPTION
			) {
			// Initialise dialog.	
			DatabaseFileChooser fc = new DatabaseFileChooser();
			
			// Remove AcceptAll file filter.
			fc.removeChoosableFileFilter(fc.getAcceptAllFileFilter());
			
			// Add database file filter.
			fc.addChoosableFileFilter(new GenBankFileFilter());
		
			// Set title.
			fc.setDialogTitle("Open database file");
			
			try {
				
				// Open file selected by user.
				if (fc.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
					return fc.getSelectedDatabaseFile();
					} 
				
				} 
			
			catch(FileTypeException fte) {
				
				JOptionPane.showMessageDialog(
					_owner, 
					"Cannot open file - has an unrecognised suffix", 
					"File suffix error", 
					JOptionPane.ERROR_MESSAGE
					);
				}
			}

		return null;
		
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	private class MyFastaFileHandler implements FastaFileHandler {
		
		private final DatabaseFile file;
		private Record record = null;
		
		public MyFastaFileHandler(final DatabaseFile file) {
			this.file = file;
			}
		
		public void header(
			final String accession, 
			final String text
			) throws DatabaseParserException, RecordConflictException {
			
			try {
				record = _database.getRecord(file, accession);	
				} 
			
			catch(IllegalArgumentException iae) {
				JOptionPane.showMessageDialog(
					_owner, 
					"Error encountered whilst parsing " + file.getName() + ".", 
					"Illegal Argument Exception", 
					JOptionPane.ERROR_MESSAGE
					);
				iae.printStackTrace();
				}	
			catch(Exception e) {
				// Unexpected exception.
				System.out.println("Unexpected exception: " + e.getMessage());
				e.printStackTrace();
				}
			
				
			}		
		
		
		public void sequence(String sequence) throws DatabaseParserException {
			
			search(record, sequence);
			
			}
		
		// NOT IMPLEMENTED.
		public void recordEnd() throws DatabaseParserException {}		
		public void recordStart() throws DatabaseParserException {}
		
		} // end of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class MyGenBankFileHandler implements GenBankFileHandler {

		private final DatabaseFile file;
		private Record record = null;
		
		public MyGenBankFileHandler(final DatabaseFile file) {
			this.file = file;
			}
		
		public void locus(
			final String locusName, 
			final int sequenceLength, 
			final MoleculeType moleculeType, 
			final GenBankDivision genBankDivision, 
			final Date date
			) throws DatabaseParserException, RecordConflictException {
			
			try {
				record = _database.getRecord(file, locusName);	
				} 
			
			catch(IllegalArgumentException iae) {
				JOptionPane.showMessageDialog(
					_owner, 
					"Error encountered whilst parsing " + file.getName() + ".", 
					"Illegal Argument Exception", 
					JOptionPane.ERROR_MESSAGE
					);
				iae.printStackTrace();
				}	
			catch(Exception e) {
				// Unexpected exception.
				System.out.println("Unexpected exception: " + e.getMessage());
				e.printStackTrace();
				}
				
			}

		
		public void origin(final String sequence) throws DatabaseParserException {
			search(record, sequence);
			}
		
		// NOT IMPLEMENTED.
		public void accession(String accession) throws DatabaseParserException, RecordConflictException {}
		public void authors(String[] authors) throws DatabaseParserException {}
		public void baseCount(int aCount, int cCount, int gCount, int uCount, int tCount, int othersCount) throws DatabaseParserException {}
		public void comment(String text) throws DatabaseParserException {}
		public void definition(String text) throws DatabaseParserException {}
		public void features(String text) throws DatabaseParserException {}
		public void featuresCDS(String text) throws DatabaseParserException {}
		public void featuresGene(String text) throws DatabaseParserException {}
		public void featuresSource(String text) throws DatabaseParserException {}
		public void journal(String text) throws DatabaseParserException {}
		public void keywords(String text) throws DatabaseParserException {}
		public void medline(String text) throws DatabaseParserException {}
		public void organism(String text) throws DatabaseParserException {}
		public void reference(String text) throws DatabaseParserException {}
		public void source(String text) throws DatabaseParserException {}
		public void title(String text) throws DatabaseParserException {}
		public void version(String version) throws DatabaseParserException {}
		
		} // End of inner class.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////