// SearchDatabaseDialog.java created on 16 May 2004 at 07:45

//=============================================================================
//
//    Copyright (C) 2004 Kevin Ashelford
//    Contact details:
//    Email: ashelford@cardiff.ac.uk
//    Address:  Cardiff School of Bioscience, Cardiff University, PO. Box 915,
//    Cardiff, UK. CF10 3TL
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//=============================================================================

package kea.programs.jPrimrose;

import java.lang.*;
import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import java.awt.event.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.util.*;

import kea.database.*;
import kea.programs.database.*;
import kea.programs.dialogBoxes.*;
import kea.programs.progress.*;
import kea.programs.buttons.*;
import kea.programs.icons.KeaIcon;
import kea.graphics.Colour;
import kea.sequences.nucleicAcid.search.*;
import kea.event.*;
import kea.programs.table.*;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/**
 Dialog for searching database with identified oligos.
 @author  Kevin Ashelford.
 @version 2.17.
 */
public final class SearchDatabaseDialog extends JDialog {
	
	//##########################################################################
	
	// CLASS FIELDS
	
	////////////////////////////////////////////////////////////////////////////
	
	// MEMBER FIELDS
	
	private final Database _database;
	private final PrimroseOligoDatabase _oligoDatabase;
	private final TargetCollection _targetCollection;
	private final OligoDesignInfo _oligoDesignInfo;
	
	private JMenu _helpMenu;
	private DialogResult _dialogResult;
	private final ProgressComponent _progressComponent;
	private JLabel _progressMessage;
	private KeaButton _searchButton;
	private Action _searchAction;
	private Action _abortAction;
	
	private Action _helpAction;
	private Action _aboutAction;
	
	private JTable _resultTable;
	private SearchDatabaseDialogTableModel _resultsTableModel;
	
	private JSpinner _nonTargetSpinner;
	
	private boolean _abortParser;
	
	private final Action _notepadAction;
	private final Action _openDatabaseAction;
	
	private DatabaseTree _databaseTree;
	
	//##########################################################################
	
	// CONSTRUCTOR
	
	/**
	 Creates a new instance of <code>SearchDatabaseDialog</code>.
	 @param owner MainClass object.
	 @param database Stores database records to search.
	 @param oligoDatabase Stores identified oligos.
	 @param targetCollection Stores target info for display only.
	 @param oligoDesignInfo User-specified information for designing oligos.
	 */
	public SearchDatabaseDialog(
		final JFrame owner,
		final Database database,
		final PrimroseOligoDatabase oligoDatabase,
		final TargetCollection targetCollection,
		final OligoDesignInfo oligoDesignInfo,
		final Action notepadAction
		) {
		
		super(owner, "Search database", true);	
		
		_notepadAction = notepadAction;
		
		_database = database;
		_oligoDatabase = oligoDatabase;
		_targetCollection = targetCollection;
		_oligoDesignInfo = oligoDesignInfo;
		
		_abortParser = true;
		
		// Initialise now for OpenDatabaseAction.
		_progressComponent = new ProgressComponent(true); 
		
		// For adding additional sequence files to database:
		_openDatabaseAction = new OpenDatabaseAction(
			this, 
			_database, 
			_progressComponent, 
			false, // Don't clear database when new files opened.
			false, // Don't show mnemonic etc.
			"Add Sequences to Database",
			"Add additional sequence files to existing database.", 
			KeaIcon.OpenDatabaseFile.getImageIcon()
			);
		
		initialise(); 
		//pack();
		this.setSize(650, 650);
		kea.util.GUIUtils.setLocationOnceSized(this);
		kea.util.GUIUtils.setWindowMinimumSize(this, getSize());
			
		
		
		} // End of constructor.
	
	//##########################################################################
	
	// PROPERTY ACCESSOR METHODS
	
	//##########################################################################
	
	// METHODS
	
	/**
	 Displays dialog.
	 @return Returns DialogResult enum according to button pressed by user.
	 */
	public DialogResult showDialog() {
		
		setVisible(true);
		return _dialogResult;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void initialise() {
		
		// Menu bar.
		setJMenuBar(menuBarComponent());
		
		JPanel mainPanel = (JPanel)getContentPane();
		mainPanel.setLayout(new BorderLayout());
			
		JPanel innerPanel1 = new JPanel(new BorderLayout());
		JPanel innerPanel2 = new JPanel(new BorderLayout());
		
		
		// Main panel north - toolbar.
		mainPanel.add(initialiseToolBarComponent(), BorderLayout.NORTH);
		
		// Main panel center - inner panel 1.
		mainPanel.add(innerPanel1, BorderLayout.CENTER);
		
		// Main panel south - progress bar.
		mainPanel.add(initialiseProgressPanel(), BorderLayout.SOUTH);
		
		
		// Inner panel 1 north - inner panel 2.
		innerPanel1.add(innerPanel2, BorderLayout.NORTH);
		
		// Inner panel 1 center - results table.
		innerPanel1.add(initialiseResultsTablePanel(), BorderLayout.CENTER);
		
		// Inner panel 1 south
		
		
		// Inner panel 2 north
		
		// Inner panel 2 center - information panel.
		innerPanel2.add(initialiseInformationPanel(), BorderLayout.CENTER);
		
		// Inner panel 2 south - button panel.
		innerPanel2.add(initialiseButtonPanel(), BorderLayout.SOUTH);
	
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JToolBar initialiseToolBarComponent() {
		
		JToolBar toolBar;
		
		toolBar = new JToolBar();
		toolBar.setRollover(true);
		toolBar.add(_openDatabaseAction);
		toolBar.addSeparator();
		toolBar.add(_notepadAction);
		toolBar.addSeparator();
		
		return toolBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseInformationPanel() {
		
		// Initialise overall panel.
		JPanel panel = new JPanel(new GridLayout());
		panel.setBorder(
			new CompoundBorder(
				new EmptyBorder(5, 5, 5, 5),
				new TitledBorder("Summary of choices made so far")
				)
			);

		// Initialise database tree object.
		_databaseTree = new DatabaseTree(_database); 
		
		// Size database tree.
		_databaseTree.setPreferredSize(new Dimension(300, 200)); // <+++++++++++++++++ SIZING
		
		//databaseTree.enableTreeRecordSelection(false);
		
		// Change cell renderer to display target icons.
		_databaseTree.getTree().setCellRenderer(
			new PrimroseTreeCellRenderer()
			);
		
		// Initialise inner panel to appear to the right of database tree.
		// Inner panel to contain secondary split pane.
		JPanel innerPanel = new JPanel(new GridLayout());
		
		// Initialise table for displaying target records used to design oligos.
		// Initialise text box for summarizing oligos identified by previous
		// dialog.
		// Initialise secondary split pane to contain table above, and text
		// box below.
		JSplitPane secondarySplitPane =
			new JSplitPane(
				JSplitPane.VERTICAL_SPLIT,
				getSelectedTargetsTablePanel(), 
				getOligoDesignInfoPanel()
				);
		innerPanel.add(secondarySplitPane);
		secondarySplitPane.setDividerLocation(130);
		secondarySplitPane.setDividerSize(7);
		secondarySplitPane.setOneTouchExpandable(true);
		secondarySplitPane.setBorder(new EmptyBorder(0,0,0,0));
		
		
		// Initialise main split pane and add database tree to left and inner 
		// panel to the right.
		JSplitPane mainSplitPane = 
			new JSplitPane(
				JSplitPane.HORIZONTAL_SPLIT, 
				_databaseTree, 
				innerPanel
				);
		panel.add(mainSplitPane);
		mainSplitPane.setDividerLocation(250);
		mainSplitPane.setDividerSize(7);
		mainSplitPane.setOneTouchExpandable(true);
		mainSplitPane.setBorder(new EmptyBorder(0,0,0,0));

		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel getOligoDesignInfoPanel() {
		
		JPanel panel = new JPanel(new GridLayout());
		
		panel.setBorder(
			new CompoundBorder(
				new EmptyBorder(0, 0, 0, 0),
				new TitledBorder("Design options selected by user")
				)
			);
		
		String message = 
			_oligoDesignInfo.getTotalOligos() + " oligonucleotides identified.\n" + 
			"Maximum number of degenerate bases allowed: " + 
			_oligoDesignInfo.getMaximumDegeneracy() + "\n" + 
			"Oligonucleotide length: " + 
			_oligoDesignInfo.getOligoLength();
		
		
		
		JTextArea textArea = new JTextArea(message);
		
		JScrollPane scrollPane = new JScrollPane(textArea);
		panel.add(scrollPane);
		panel.setPreferredSize(new Dimension(150, 70));  //<+++++++++++++++++++++++ SIZING
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel getSelectedTargetsTablePanel() {
		
		JPanel panel = new JPanel(new GridLayout());
		
		panel.setBorder(
			new CompoundBorder(
				new EmptyBorder(0, 0, 0, 0),
				new TitledBorder("Records used to design oligonucleotides")
				)
			);
		
		// Get selected targets.
		final Target[] selectedTargets = _targetCollection.getSelectedTargets();
		
		// Initialse table model.
		final TableModel model = new SelectedTargetsTableModel(selectedTargets);
		
		// Initialise table.
		final JTable table = new JTable(model);
		
		// Initialise TableCellRenderer.
		TableCellRenderer renderer = new FindOligoDialogTableCellRenderer();
		
		// Add TableCellRenderer for dealing with name and icon objects.
		table.setDefaultRenderer(Record.class, renderer);
		
		// Determine how columns should be resized.
		table.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
		
		// Set column sizes
		TableColumn column = null;
		for (int i = 0; i < 3; i++) {
			column = table.getColumnModel().getColumn(i);
			if (i == 0) { // Number
				column.setPreferredWidth(180); 	
				} 
			else {
				column.setPreferredWidth(50);
				}
			}
		
		// Do not display grid.
		table.setShowGrid(false);
		
		// Set selection colours.
		table.setSelectionBackground(Colour.BackgroundSelection);
		table.setSelectionForeground(Color.BLACK);
		
		// Add to scroll pane, set background to white and return.
		JScrollPane sp = new JScrollPane(table);
		sp.getViewport().setBackground(Color.WHITE);
		
		panel.add(sp);
		
		panel.setPreferredSize(new Dimension(150, 150)); // <+++++++++++++++++++++++ KEY SIZING STEP!
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseButtonPanel() {
		
		_searchAction = new SearchAction();
		_abortAction = new AbortAction();
		_abortAction.setEnabled(false);
		
		JPanel panel = new JPanel(new GridLayout(1, 3, 5, 5));
		panel.setBorder(new EmptyBorder(5, 5, 5, 5));
		
		JPanel parameterPanel = new JPanel(new GridLayout(1, 1, 0, 0));
		panel.add(parameterPanel);
		
		_searchButton = new KeaButton(
			_searchAction, 
			KeaButtonColour.Green
			);
		panel.add(_searchButton);
		
		KeaButton abortButton = new KeaButton(
			_abortAction, 
			KeaButtonColour.Red
			);
		panel.add(abortButton);
		
		JLabel nonTargetSpinnerLabel = new JLabel("Maximum non-targets: ");
		nonTargetSpinnerLabel.setHorizontalAlignment(JLabel.RIGHT);
		
		_nonTargetSpinner = new JSpinner(
			new SpinnerNumberModel(	
				0,					// Value
				0,					// Min 
				_database.getSize() - _targetCollection.getSize(),// Max
				1					// step
				)
			);
		((JSpinner.DefaultEditor)_nonTargetSpinner.
			getEditor()).
				getTextField().
					setColumns(5);
		
		JPanel nonTargetSpinnerPanel = new JPanel(new BorderLayout());
		
		nonTargetSpinnerPanel.add(nonTargetSpinnerLabel, BorderLayout.CENTER);
		nonTargetSpinnerPanel.add(_nonTargetSpinner, BorderLayout.EAST);
		
		parameterPanel.add(nonTargetSpinnerPanel);
		
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseResultsTablePanel() {
		
		JPanel panel = new JPanel(new GridLayout(1, 1));
		panel.setBorder(
			new CompoundBorder(
				new EmptyBorder(5, 5, 5, 5),
				new TitledBorder("Results of search")
				)
			);
		
		// Initialise TableModel with empty collection.
		_resultsTableModel = 
			new SearchDatabaseDialogTableModel(new PrimroseOligo[0]);
		
		// Initialise table with TableModel.
		//_resultTable = new JTable(_resultsTableModel);  // not used, see below.
		
		// Enable sorting - uses classes from Java tutorial.
		TableSorter sorter = new TableSorter(_resultsTableModel); 
		_resultTable = new JTable(sorter);           
		sorter.addMouseListenerToHeaderInTable(_resultTable); 

		
		// Add TableCellRenderer for dealing with name and icon objects.
		_resultTable.setDefaultRenderer(Oligo.class, new PrimroseOligoTableCellRenderer());
		_resultTable.setDefaultRenderer(Float.class, new FloatTableCellRenderer());
		_resultTable.setDefaultRenderer(Integer.class, new IntegerTableCellRenderer());
		
		// Determine how columns should be resized.
		_resultTable.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
		
		// Size columns.
		sizeResultTableColumns();
		
		// Do not display grid.
		_resultTable.setShowGrid(true);
		
		_resultTable.setCellSelectionEnabled(true);
		
		// Set selection colours.
		_resultTable.setSelectionBackground(Colour.BackgroundSelection);
		_resultTable.setSelectionForeground(Color.BLACK);
		
		_resultTable.addMouseListener(new ResultTableMouseListener());
		
		// Add to scroll pane, set background to white and return.
		JScrollPane sp = new JScrollPane(_resultTable);
		sp.getViewport().setBackground(Color.WHITE);
		
		panel.add(sp);
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void sizeResultTableColumns() {
		
		// Number of columns:
		int total = _resultTable.getColumnModel().getColumnCount();
		
		// Set column sizes
		TableColumn column = null;
		for (int i = 0; i < total; i++) {
			column = _resultTable.getColumnModel().getColumn(i);
			if (i == 0) {
				column.setPreferredWidth(30);
				} 
			else if (i == 1) { 
				column.setPreferredWidth(180); 	
				} 
			else if (i == 2) {
				column.setPreferredWidth(30);
				}
			else if (i == 3) {
				column.setPreferredWidth(40);
				}
			else if (i == 4) {
				column.setPreferredWidth(45);
				}
			else if (i == 5) {
				column.setPreferredWidth(45);
				}
			else if (i == 6) {
				column.setPreferredWidth(45);
				}
			else if (i == 7) {
				column.setPreferredWidth(30);
				}
			else if (i == 8) {
				column.setPreferredWidth(30);
				}
			else {
				column.setPreferredWidth(50);
				}
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JPanel initialiseProgressPanel() {
		
		JPanel panel = new JPanel(new GridLayout(2, 1, 2, 2));
		panel.setBorder(new EmptyBorder(0, 2, 2, 2));
		
		_progressMessage = new JLabel();
		_progressMessage.setBorder(new TitledBorder(""));
		_progressMessage.setPreferredSize(new Dimension(150,30));
		panel.add(_progressMessage);
		
		panel.add(_progressComponent);
		
		return panel;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private JMenuBar menuBarComponent() {
		
		JMenuBar menuBar = new JMenuBar();
		menuBar.setBorder(new EmptyBorder(0,0,0,0));
		
		// File
		JMenu fileMenu = new JMenu("File");
		menuBar.add(fileMenu);
		fileMenu.getAccessibleContext().setAccessibleName("File");
		fileMenu.setMnemonic(KeyEvent.VK_F);
		fileMenu.add(_openDatabaseAction);
		
		// Help.
		_helpMenu = new JMenu("Help");
		menuBar.add(_helpMenu);
		_helpMenu.getAccessibleContext().setAccessibleName("Help");
		_helpMenu.setMnemonic(KeyEvent.VK_H);
		
		return menuBar;
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/**
	 Use this method to add menu items (Actions) to the help menu.
	 @param action Action to add.
	 */
	public void addHelpMenuAction(final Action action) {
		
		if (action.getValue(Action.NAME) == "Help") {
			_helpAction = action;
			} 
		
		if (action.getValue(Action.NAME) == "About") {
			_aboutAction = action;
			}
		
		_helpMenu.add(action);
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void search() {
		
		// Reset:
		//=================================
		_abortParser = false;
		_resultsTableModel.clear(); 
		sizeResultTableColumns();
		_progressComponent.reset();
		_progressMessage.setText("");
		resetOligoDatabase(_oligoDatabase);	
		//=================================
		
		final int allowedNonTargetHits = 
			((Integer)this._nonTargetSpinner.getValue()).intValue();
		
		final AbortListener abortListener = new AbortListener() {
			public boolean abort() {return _abortParser;} 
			};
		
		final SearchDatabase searchDatabase = new SearchDatabase(  
			this, 
			_database, 
			_oligoDatabase.getOligos(),
			_progressComponent,
			_progressMessage, 
			allowedNonTargetHits, 
			abortListener
			);
		
		_searchAction.setEnabled(false);
		_abortAction.setEnabled(true);
		_openDatabaseAction.setEnabled(false);
		
		_databaseTree.enableTreeRecordSelection(false);
		
		// Run search on separate thread.
		//===================================================================
		Thread t = new Thread(new Runnable(){public void run() {
			
			// DO SEARCH
			searchDatabase.run(); 
			
			// Successfully completed search.
			if (!_abortParser) {		
				displayResults();
				_abortParser = true;
				} 
			
			// Aborted search.
			else {
				_resultsTableModel.clear(); 
				sizeResultTableColumns();
				_progressComponent.reset();
				SwingUtilities.invokeLater(new Runnable(){public void run(){
					_progressMessage.setText("");
					}});
				resetOligoDatabase(_oligoDatabase);	
				}
			
			
			_searchAction.setEnabled(true);
			_abortAction.setEnabled(false);
			_openDatabaseAction.setEnabled(true);
			
			_databaseTree.enableTreeRecordSelection(true);
			
			}});	
		t.setPriority(t.MIN_PRIORITY);
		t.start();
		//====================================================================
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void displayResults() {
		
		ArrayList buffer = new ArrayList();
		
		// Winnow out failed oligos.
		PrimroseOligo[] originalOligos = _oligoDatabase.getOligos();
		for (int i = 0; i < originalOligos.length; i++) {
			if(!originalOligos[i].hasFailed()) {buffer.add(originalOligos[i]);}
			}
		
		PrimroseOligo[] oligos = new PrimroseOligo[buffer.size()];
		buffer.toArray(oligos);
		
		_resultsTableModel.updateModel(oligos); 
		sizeResultTableColumns();
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void resetOligoDatabase(PrimroseOligoDatabase oligoDatabase) {
		PrimroseOligo[] oligos = oligoDatabase.getOligos();
		for (int i = 0; i < oligos.length; i++) {
			oligos[i].reset();
			}
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private void abort() {
		_abortParser = true;
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	private boolean okToClose() {
		
		
		
		// No search / no data.
		if (_abortParser && _resultTable.getRowCount() == 0) {
			return true;
			}
		
		// Search running.
		else if (!_abortParser) {
			
			int userResponse = 
				JOptionPane.showConfirmDialog(
					this, 
					"<html>This will cancel your search.<p>" + 
					"Are you sure you want to do this?</html>",
					"Closing...",
					JOptionPane.YES_NO_OPTION
					);
			
			if (userResponse == JOptionPane.YES_OPTION) {
				return true;
				} 
			else {
				return false;
				}
			
			}
		
		// data displayed.
		else {
			int userResponse = 
				JOptionPane.showConfirmDialog(
					this, 
					"<html>Results will be lost if you close this window.<p>" + 
					"Do you want to continue?</html>",
					"Closing...",
					JOptionPane.YES_NO_OPTION
					);
			
			if (userResponse == JOptionPane.YES_OPTION) {
				return true;
				} 
			else {
				return false;
				}
			}
		
		} // End of method.
	
	////////////////////////////////////////////////////////////////////////////
	
	/** 
	 Overrides base class method if the WINDOW_CLOSING window event is fired.
	 @param we Event passed from EventListener.
	 */
	protected void processWindowEvent(WindowEvent we) {
		
		// Action if window closing exit is being fired.
		if (we.getID() == WindowEvent.WINDOW_CLOSING) {
			
			if (okToClose()) {
				if (!_abortParser) {abort();} // abort parser if still running.
				setVisible(false);
				dispose();
				}
			
			} 
		// Otherwise use base class method.
		else {
			super.processWindowEvent(we);
			}
		
		} // End of method.
	
	//##########################################################################
	
	// INNER CLASSES
	
	private class SearchAction extends AbstractAction {
		
		public SearchAction() {
			putValue(Action.NAME, "Search Database Records");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Search database records with identified oligonucleotides."
				);
			putValue(Action.SMALL_ICON, KeaIcon.AnalyseDatabase.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent ae) {
			search();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class AbortAction extends AbstractAction {
		
		public AbortAction() {
			putValue(Action.NAME, "Abort Search");
			putValue(
				Action.SHORT_DESCRIPTION, 
				"Cancels current database search."
				);
			putValue(Action.SMALL_ICON, KeaIcon.Cancel.getImageIcon());
			}
		
		public void actionPerformed(ActionEvent e) {
			abort();
			}
		
		} // End of inner class.
	
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	
	private class ResultTableMouseListener extends MouseAdapter {
		
		public void mouseClicked(MouseEvent me) {
			if (me.getClickCount() == 2) {display(me);}
			}
		
		// Display details of selected oligo.
		private void display(MouseEvent me) { 
			
			int rowIndex = _resultTable.rowAtPoint(me.getPoint());
			int columnIndex = _resultTable.columnAtPoint(me.getPoint());
			
			Object o = 
				_resultTable.getModel().getValueAt(rowIndex, columnIndex);
			
			if (o instanceof PrimroseOligo) {
				PrimroseOligo oligo = (PrimroseOligo)o;
				
				Report report = new Report( 
					SearchDatabaseDialog.this, 
					oligo, 
					_database
					);
				report.addHelpMenuAction(_helpAction);
				report.addHelpMenuAction(_aboutAction);
				report.showDialog();  
				}
			} 
		} // End of inner class.
	
	//##########################################################################
	
	} // End of class.

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////